import sys
import logging
import inspect
try:
    from importlib.metadata import entry_points

    if sys.version_info < (3, 10):
        _entry_points = entry_points

        def entry_points(group=None):
            return _entry_points().get(group, [])
except ImportError:
    from pkg_resources import iter_entry_points as entry_points

import glad.specification
from glad.generator.c import CGenerator
from glad.generator.rust import RustGenerator
from glad.parse import Specification


logger = logging.getLogger(__name__)


GENERATOR_ENTRY_POINT = 'glad.generator'
SPECIFICATION_ENTRY_POINT = 'glad.specification'


DEFAULT_GENERATORS = dict(
    c=CGenerator,
    rust=RustGenerator
)
DEFAULT_SPECIFICATIONS = dict()

for name, cls in inspect.getmembers(glad.specification, inspect.isclass):
    if issubclass(cls, Specification) and cls is not Specification:
        DEFAULT_SPECIFICATIONS[cls.NAME] = cls


def find_generators(default=None, entry_point=GENERATOR_ENTRY_POINT):
    generators = dict(DEFAULT_GENERATORS if default is None else default)

    for entry_point in entry_points(group=entry_point):
        generators[entry_point.name] = entry_point.load()
        logger.debug('loaded language %s: %s', entry_point.name, generators[entry_point.name])

    return generators


def find_specifications(default=None, entry_point=SPECIFICATION_ENTRY_POINT):
    specifications = dict(DEFAULT_SPECIFICATIONS if default is None else default)

    for entry_point in entry_points(group=entry_point):
        specifications[entry_point.name] = entry_point.load()
        logger.debug('loaded specification %s: %s', entry_point.name, specifications[entry_point.name])

    return specifications



from contextlib import closing
import logging
import sys

if sys.version_info >= (3, 0):
    _is_py3 = True
    from urllib.request import build_opener, ContentTooShortError
else:
    _is_py3 = False
    from urllib2 import build_opener
    from urllib import FancyURLopener


logger = logging.getLogger('glad.opener')


def build_urllib_opener(user_agent, *args, **kwargs):
    if _is_py3:
        return None

    class UrllibURLOpener(FancyURLopener):
        version = user_agent
    return UrllibURLOpener(*args, **kwargs)


def _urlretrieve_with_opener(opener, url, filename, data=None):
    if not _is_py3:
        raise SyntaxError('Only call this in Python 3 code.')

    # borrowed from the original implementation at urllib.request.urlretrieve.
    with closing(opener.open(url, data=data)) as src:
        headers = src.info()

        with open(filename, 'wb') as dest:
            result = filename, headers
            bs = 1024*8
            size = -1
            read = 0
            blocknum = 0
            if "content-length" in headers:
                size = int(headers["Content-Length"])

            while True:
                block = src.read(bs)
                if not block:
                    break
                read += len(block)
                dest.write(block)
                blocknum += 1

    if size >= 0 and read < size:
        raise ContentTooShortError(
            'retrieval incomplete: got only %i out of %i bytes'
            % (read, size), result)

    return result


class URLOpener(object):
    """
    Class to download/find Khronos related files, like
    the official specs and khrplatform.h.

    Can also be used to download files, exists mainly because of
    Python 2 and Python 3 incompatibilities.
    """
    def __init__(self, user_agent='Mozilla/5.0'):
        # the urllib2/urllib.request opener
        self.opener = build_opener()
        self.opener.addheaders = [('User-agent', user_agent)]

        # the urllib opener (Python 2 only)
        self.opener2 = build_urllib_opener(user_agent)

    def urlopen(self, url, data=None, *args, **kwargs):
        """
        Same as urllib2.urlopen or urllib.request.urlopen,
        the only difference is that it links to the internal opener.
        """
        logger.info('opening: \'%s\'', url)

        if data is None:
            return self.opener.open(url)

        return self.opener.open(url, data)

    def urlretrieve(self, url, filename, data=None):
        """
        Similar to urllib.urlretrieve or urllib.request.urlretrieve
        only that *filname* is required.

        :param url: URL to download.
        :param filename: Filename to save the content to.
        :param data: Valid URL-encoded data.
        :return: Tuple containing path and headers.
        """
        logger.info('saving: \'%s\' to \'%s\'', url, filename)

        if _is_py3:
            return _urlretrieve_with_opener(self.opener, url, filename, data=data)

        return self.opener2.retrieve(url, filename, data=data)

    # just a singleton helper:
    _default = None

    @classmethod
    def default(cls):
        if cls._default is None:
            cls._default = cls()

        return cls._default



import functools
import os
import re
import sys
from collections import namedtuple, defaultdict


if sys.version_info >= (3, 0, 0):
    basestring = str


Version = namedtuple('Version', ['major', 'minor'])
ExpandedName = namedtuple('ExpandedName', ['prefix', 'suffix'])


_API_NAMES = {
    'egl': 'EGL',
    'gl': 'OpenGL',
    'gles1': 'OpenGL ES',
    'gles2': 'OpenGL ES',
    'glsc2': 'OpenGL SC',
    'glx': 'GLX',
    'wgl': 'WGL',
}


def api_name(api):
    api = api.lower()
    return _API_NAMES.get(api, api.upper())


def makefiledir(path):
    dir = os.path.split(path)[0]
    if not os.path.exists(dir):
        os.makedirs(dir)


ApiInformation = namedtuple('ApiInformation', ('specification', 'version', 'profile'))
_API_SPEC_MAPPING = {
    'gl': 'gl',
    'gles1': 'gl',
    'gles2': 'gl',
    'glsc2': 'gl',
    'egl': 'egl',
    'glx': 'glx',
    'wgl': 'wgl',
    'vulkan': 'vk'
}


def parse_version(value):
    if value is None:
        return None

    value = value.strip()
    if not value:
        return None

    major, minor = (value + '.0').split('.')[:2]
    return Version(int(major), int(minor))


def parse_apis(value, api_spec_mapping=_API_SPEC_MAPPING):
    result = dict()

    for api in value.split(','):
        api = api.strip()

        m = re.match(
            r'^(?P<api>\w+)(:(?P<profile>\w+))?(/(?P<spec>\w+))?(=(?P<version>\d+(\.\d+)?)?)?$',
            api
        )

        if m is None:
            raise ValueError('Invalid API {}'.format(api))

        spec = m.group('spec')
        if spec is None:
            try:
                spec = api_spec_mapping[m.group('api')]
            except KeyError:
                raise ValueError('Can not resolve specification for API {}'.format(m.group('api')))

        version = parse_version(m.group('version'))

        result[m.group('api')] = ApiInformation(spec, version, m.group('profile'))

    return result


# based on https://stackoverflow.com/a/11564323/969534
def topological_sort(items, key, dependencies):
    pending = [(item, set(dependencies(item))) for item in items]
    emitted = []
    while pending:
        next_pending = []
        next_emitted = []
        for entry in pending:
            item, deps = entry
            deps.difference_update(emitted)
            if deps:
                next_pending.append(entry)
            else:
                yield item
                key_item = key(item)
                emitted.append(key_item)
                next_emitted.append(key_item)
        if not next_emitted:
            raise ValueError("cyclic or missing dependency detected: %r" % (next_pending,))
        pending = next_pending
        emitted = next_emitted


class _HashedSeq(list):
    __slots__ = 'hashvalue'

    # noinspection PyMissingConstructor
    def __init__(self, tup, hash=hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue


def _default_key_func(*args, **kwargs):
    key = (tuple(args), tuple(kwargs.items()))
    return _HashedSeq(key)


def memoize(key=None, method=False):
    """
    Memoize decorator for functions and methods.

    :param key: a cache-key transformation function
    :param method: whether the cache should be attached to the `self` parameter
    """
    key_func = key or _default_key_func

    def memoize_decorator(func):
        _cache = dict()

        @functools.wraps(func)
        def memoized(*args, **kwargs):
            cache_args = args
            if method:
                # This is an attempt to bind the cache to the instance of the currently
                # executed method. The idea is to not hoard references to the instance
                # and other values (arguments) to not prevent the GC from collecting those.
                # If we don't attach it this leaks memory all over the place,
                # especially since this implementation currently has an uncapped cache.
                self = args[0]
                cache_args = args[1:]
                try:
                    funcs_cache = self._memoize_cache
                except AttributeError:
                    funcs_cache = defaultdict(dict)
                    self._memoize_cache = funcs_cache
                cache = funcs_cache[func]
            else:
                cache = _cache

            key = key_func(*cache_args, **kwargs)
            if key not in cache:
                cache[key] = func(*args, **kwargs)
            return cache[key]

        return memoized

    return memoize_decorator


def itertext(element, ignore=()):
    tag = element.tag
    if not isinstance(tag, basestring) and tag is not None:
        return
    if element.text:
        yield element.text
    for e in element:
        if not e.tag in ignore:
            for s in itertext(e, ignore=ignore):
                yield s
            if e.tail:
                yield e.tail


def expand_type_name(name):
    """
    Transforms a type name into its expanded version, e.g.
    expands the type `VkShaderInfoTypeAMD` to the tuple `('VK_SHADER_INFO_TYPE', '_AMD')`.

    See: https://github.com/KhronosGroup/Vulkan-Docs/blob/main/scripts/generator.py#L60 buildEnumCDecl_Enum
    """
    upper_name = re.sub(r'([0-9]+|[a-z_])([A-Z0-9])', r'\1_\2', name).upper()
    (prefix, suffix) = (upper_name, '')

    suffix_match = re.search(r'[A-Z][A-Z]+$', name)
    if suffix_match:
        suffix = '_' + suffix_match.group()
        # Strip off the suffix from the prefix
        prefix = upper_name.rsplit(suffix, 1)[0]

    return ExpandedName(prefix, suffix)



import collections
import logging


class Sink(object):
    def info(self, message, exc=None):
        raise NotImplementedError

    def warning(self, message, exc=None):
        raise NotImplementedError

    def error(self, message, exc=None):
        raise NotImplementedError


class NullSink(Sink):
    def info(self, message, exc=None):
        pass

    def warning(self, message, exc=None):
        pass

    def error(self, message, exc=None):
        pass


class LoggingSink(Sink):
    _DEFAULT_LOGGER = logging.getLogger(__name__)

    def __init__(self, name=None, logger=None):
        self.logger = logger
        if self.logger is None and name is not None:
            self.logger = logging.getLogger(name)
        if self.logger is None:
            self.logger = self._DEFAULT_LOGGER

    def info(self, message, exc=None):
        self.logger.info(message)

    def warning(self, message, exc=None):
        self.logger.warning(message)

    def error(self, message, exc=None):
        self.logger.error(message)


Message = collections.namedtuple('Message', ['type', 'content', 'exc'])


class CollectingSink(Sink):
    def __init__(self):
        self.messages = list()

    def info(self, message, exc=None):
        self.messages.append(Message('info', message, exc))

    def warning(self, message, exc=None):
        self.messages.append(Message('warning', message, exc))

    def error(self, message, exc=None):
        self.messages.append(Message('error', message, exc))

    @property
    def infos(self):
        return [m for m in self.messages if m.type == 'info']

    @property
    def warnings(self):
        return [m for m in self.messages if m.type == 'warning']

    @property
    def errors(self):
        return [m for m in self.messages if m.type == 'error']




import jinja2

import glad
from glad.config import Config, ConfigOption
from glad.generator import JinjaGenerator
from glad.generator.util import (
    strip_specification_prefix,
    collect_alias_information,
    find_extensions_with_aliases,
    jinja2_contextfilter
)
from glad.parse import ParsedType, EnumType
from glad.sink import LoggingSink


_RUST_TYPE_MAPPING = {
    'void': 'c_void',
    'char': 'c_char',
    'uchar': 'c_uchar',
    'float': 'c_float',
    'double': 'c_double',
    'int': 'c_int',
    'long': 'c_long',
    'int8_t': 'i8',
    'uint8_t': 'u8',
    'int16_t': 'i16',
    'uint16_t': 'u16',
    'int32_t': 'i32',
    'int64_t': 'i32',
    'uint32_t': 'u32',
    'uint64_t': 'u64',
    'size_t': 'usize',
    'ull': 'u64',
}


def enum_type(enum, feature_set):
    if enum.alias and enum.value is None:
        aliased = feature_set.find_enum(enum.alias)
        if aliased is None:
            raise ValueError('unable to resolve enum alias {} of enum {}'.format(enum.alias, enum))
        enum = aliased

    # if the value links to another enum, resolve the enum right now
    if enum.value is not None:
        # enum = feature_set.find_enum(enum.value, default=enum)
        referenced = feature_set.find_enum(enum.value)
        # TODO currently every enum with a parent type is u32
        if referenced is not None and referenced.parent_type is not None:
            return 'u32'

    # we could return GLenum and friends here
    # but thanks to type aliasing we don't have to
    # this makes handling types for different specifications
    # easier, since we don't have to swap types. GLenum -> XXenum.
    if enum.type:
        _RUST_TYPE_MAPPING.get(enum.type, 'c_uint')

    if enum.value.startswith('0x'):
        return 'u64' if len(enum.value[2:]) > 8 else 'c_uint'

    if enum.name in ('GL_TRUE', 'GL_FALSE'):
        return 'c_uchar'

    if enum.value.startswith('-'):
        return 'c_int'

    if enum.value.endswith('f') or enum.value.endswith('F'):
        return 'c_float'

    if enum.value.startswith('"'):
        # TODO figure out correct type
        return '&str'

    if enum.value.startswith('(('):
        # Casts: '((Type)value)' -> 'Type'
        raise NotImplementedError

    if enum.value.startswith('EGL_CAST'):
        # EGL_CAST(type,value) -> type
        return enum.value.split('(', 1)[1].split(',')[0]

    return 'c_uint'


def enum_value(enum, feature_set):
    if enum.alias and enum.value is None:
        enum = feature_set.find_enum(enum.alias)

    # basically an alias to another enum (value contains another enum)
    # resolve it here and adjust value accordingly.
    referenced = feature_set.find_enum(enum.value)
    if referenced is None:
        pass
    elif referenced.parent_type is not None:
        # global value is a reference to a enum type value
        return '{}::{} as u32'.format(referenced.parent_type, enum.value)
    else:
        enum = referenced

    value = enum.value
    if value.endswith('"'):
        value = value[:-1] + r'\0"'
        return value

    if enum.value.startswith('EGL_CAST'):
        # EGL_CAST(type,value) -> value as type
        type_, value = enum.value.split('(', 1)[1].rsplit(')', 1)[0].split(',')
        return '{} as {}'.format(value, type_)

    if enum.type == 'float' and value.endswith('F'):
        value = value[:-1]

    for old, new in (('(', ''), (')', ''), ('f', ''),
                     ('U', ''), ('L', ''), ('~', '!')):
        value = value.replace(old, new)

    return value


def to_rust_type(type_):
    if type_ is None:
        return 'std::os::raw::c_void'

    parsed_type = type_ if isinstance(type_, ParsedType) else ParsedType.from_string(type_)

    if not parsed_type.is_pointer and parsed_type.type == 'void':
        return '()'

    prefix = ''
    if parsed_type.is_pointer > 0:
        if parsed_type.is_const:
            prefix = '*const ' * parsed_type.is_pointer
        else:
            prefix = '*mut ' * parsed_type.is_pointer

    type_ = _RUST_TYPE_MAPPING.get(parsed_type.type, parsed_type.type)

    if parsed_type.is_array > 0:
        type_ = '[{};{}]'.format(type_, parsed_type.is_array)

    return ' '.join(e.strip() for e in (prefix, type_)).strip()


def to_rust_params(command, mode='full'):
    if mode == 'names':
        return ', '.join(identifier(param.name) for param in command.params)
    elif mode == 'types':
        return ', '.join(to_rust_type(param.type) for param in command.params)
    elif mode == 'full':
        return ', '.join(
            '{name}: {type}'.format(name=identifier(param.name), type=to_rust_type(param.type))
            for param in command.params
        )

    raise ValueError('invalid mode: ' + mode)


def identifier(name):
    if name in ('type', 'ref', 'box', 'in'):
        return name + '_'
    return name


class RustConfig(Config):
    ALIAS = ConfigOption(
        converter=bool,
        default=False,
        description='Automatically adds all extensions that ' +
                    'provide aliases for the current feature set.'
    )
    MX = ConfigOption(
        converter=bool,
        default=False,
        description='Enables support for multiple GL contexts'
    )


class RustGenerator(JinjaGenerator):
    DISPLAY_NAME = 'Rust'

    TEMPLATES = ['glad.generator.rust']
    Config = RustConfig

    def __init__(self, *args, **kwargs):
        JinjaGenerator.__init__(self, *args, **kwargs)

        self.environment.filters.update(
            feature=lambda x: 'feature = "{}"'.format(x),
            enum_type=jinja2_contextfilter(lambda ctx, enum: enum_type(enum, ctx['feature_set'])),
            enum_value=jinja2_contextfilter(lambda ctx, enum: enum_value(enum, ctx['feature_set'])),
            type=to_rust_type,
            params=to_rust_params,
            identifier=identifier,
            no_prefix=jinja2_contextfilter(lambda ctx, value: strip_specification_prefix(value, ctx['spec']))
        )

    @property
    def id(self):
        return 'rust'

    def select(self, spec, api, version, profile, extensions, config, sink=LoggingSink(__name__)):
        if extensions is not None:
            extensions = set(extensions)

            if config['ALIAS']:
                extensions.update(find_extensions_with_aliases(spec, api, version, profile, extensions))

        return JinjaGenerator.select(self, spec, api, version, profile, extensions, config, sink=sink)

    def get_template_arguments(self, spec, feature_set, config):
        args = JinjaGenerator.get_template_arguments(self, spec, feature_set, config)

        args.update(
            version=glad.__version__,
            aliases=collect_alias_information(feature_set.commands)
        )

        return args

    def get_templates(self, spec, feature_set, config):
        return [
            ('Cargo.toml', 'glad-{}/Cargo.toml'.format(feature_set.name)),
            ('lib.rs', 'glad-{}/src/lib.rs'.format(feature_set.name)),
            ('impl.rs', 'glad-{}/src/{}.rs'.format(feature_set.name, spec.name))
        ]

    def modify_feature_set(self, spec, feature_set, config):
        self._remove_empty_enums(feature_set)

        return feature_set

    def _remove_empty_enums(self, feature_set):
        """
        There are some enums which are simply empty:
        https://github.com/KhronosGroup/Vulkan-Docs/issues/1754
        they need to be removed, we need to also remove any type which is an alias to that empty enum.
        """
        to_remove = set()

        for typ in (t for t in feature_set.types if isinstance(t, EnumType)):
            if typ.alias is None and not typ.enums_for(feature_set):
                to_remove.add(typ.name)

        feature_set.types = [t for t in feature_set.types if t.name not in to_remove and t.alias not in to_remove]



#!/usr/bin/env python

"""
Uses the official Khronos-XML specs to generate a
Vulkan/GL/GLES/EGL/GLX/WGL Loader made for your needs.


Subcommands have additional help information, query with: `{subcommand} --help`
"""
from itertools import groupby

import logging
import os

import glad.files
from glad import __version__
from glad.config import Config, ConfigOption
from glad.generator import GenerationInfo
from glad.sink import LoggingSink
from glad.opener import URLOpener
from glad.parse import FeatureSet
from glad.plugin import find_specifications, find_generators
from glad.util import parse_apis


logger = logging.getLogger('glad')


def parse_extensions(value):
    if os.path.isfile(value):
        # it's an extensions file
        with open(value) as f:
            value = f.read()

    value = value.replace(',', ' ')
    return list(filter(None, value.split()))


class GlobalConfig(Config):
    OUT_PATH = ConfigOption(
        required=True,
        description='Output directory for the generated files'
    )
    API = ConfigOption(
        required=True,
        converter=parse_apis,
        description='Comma separated list of APIs in `name:profile=version` pairs '
                    'optionally including a specification `name:profile/spec=version`. '
                    'No version means latest, a profile is only required if the API requires a profile. '
                    'E.g. `gl:core=3.3,gles1/gl=2,gles2'
    )
    EXTENSIONS = ConfigOption(
        converter=parse_extensions,
        default=None,
        description='Path to a file containing a list of extensions or '
                    'a comma separated list of extensions, if missing '
                    'all possible extensions are included.'
    )
    MERGE = ConfigOption(
        converter=bool,
        default=False,
        description='Merge multiple APIs of the same specification into one file.'
    )
    QUIET = ConfigOption(
        converter=bool,
        description='Disable logging.'
    )
    REPRODUCIBLE = ConfigOption(
        converter=bool,
        default=False,
        description='Makes the build reproducible by not fetching the latest '
                    'specification from Khronos.'
    )


def load_specifications(specification_names, opener, specification_classes=None):
    specifications = dict()

    if specification_classes is None:
        specification_classes = find_specifications()

    for name in set(specification_names):
        Specification = specification_classes[name]
        xml_name = name + '.xml'

        if os.path.isfile(xml_name):
            logger.info('using local specification: %s', xml_name)
            specification = Specification.from_file(xml_name, opener=opener)
        else:
            logger.info('getting %r specification from remote location', name)
            specification = Specification.from_remote(opener=opener)

        specifications[name] = specification

    return specifications


def apis_by_specification(api_info, specifications):
    return groupby(api_info.items(),
                   key=lambda api_info: specifications[api_info[1].specification])


def main(args=None):
    from argparse import ArgumentParser
    import sys

    # Initialize logging as early as possible
    if not '--quiet' in (args or sys.argv):
        logging.basicConfig(
            format='[%(asctime)s][%(levelname)s\t][%(name)-7s\t]: %(message)s',
            datefmt='%d.%m.%Y %H:%M:%S', level=logging.DEBUG
        )

    logging_sink = LoggingSink(logger=logger)

    description = __doc__
    parser = ArgumentParser(description=description)

    parser.add_argument('--version', action='version', version=__version__)

    global_config = GlobalConfig()
    global_config.init_parser(parser)

    subparsers = parser.add_subparsers(
        dest='subparser_name',
        description='Generator to use'
    )
    subparsers.default = 'c'

    configs = dict()
    generators = find_generators()
    for lang, Generator in generators.items():
        config = Generator.Config()
        subparser = subparsers.add_parser(lang)
        config.init_parser(subparser)

        configs[lang] = config

    ns = parser.parse_args(args=args)

    global_config.update_from_object(ns, convert=False, ignore_additional=True)
    config = configs[ns.subparser_name]
    config.update_from_object(ns, convert=False, ignore_additional=True)

    # This should never throw if Config.init_parser is working correctly
    global_config.validate()  # Done before, but doesn't hurt
    config.validate()

    if global_config['REPRODUCIBLE']:
        opener = glad.files.StaticFileOpener()
        gen_info_factory = lambda *a, **kw: GenerationInfo.create(when='-', *a, **kw)
    else:
        opener = URLOpener()
        gen_info_factory = GenerationInfo.create

    specifications = load_specifications(
        [value[0] for value in global_config['API'].values()], opener=opener
    )

    generator = generators[ns.subparser_name](
        global_config['OUT_PATH'], opener=opener, gen_info_factory=gen_info_factory
    )

    invalid_extensions = set(global_config['EXTENSIONS'] or [])
    for specification, apis in apis_by_specification(global_config['API'], specifications):
        for api in apis:
            invalid_extensions = invalid_extensions.difference(specification.extensions[api[0]])

    if not len(invalid_extensions) == 0:
        message = 'invalid extensions or extensions not present in one of the selected APIs: {}\n' \
            .format(', '.join(invalid_extensions))
        parser.exit(11, message)

    def select(specification, api, info):
        logging_sink.info('generating {}:{}/{}={}'.format(api, info.profile, info.specification, info.version))

        extensions = global_config['EXTENSIONS']
        if extensions:
            extensions = [ext for ext in extensions if specification.is_extension(api, ext)]

        return generator.select(specification, api, info.version, info.profile, extensions, config, sink=logging_sink)

    for specification, apis in apis_by_specification(global_config['API'], specifications):
        feature_sets = list(select(specification, api, info) for api, info in apis)

        if global_config['MERGE']:
            logging_sink.info('merging {}'.format(feature_sets))
            feature_sets = [FeatureSet.merge(feature_sets, sink=logging_sink)]
            logging_sink.info('merged into {}'.format(feature_sets[0]))

        for feature_set in feature_sets:
            logging_sink.info('generating feature set {}'.format(feature_set))
            generator.generate(specification, feature_set, config, sink=logging_sink)


if __name__ == '__main__':
    main()



from datetime import datetime

import sys

import collections
import os.path
from jinja2 import Environment, ChoiceLoader, PackageLoader

import glad
from glad.config import Config
from glad.sink import LoggingSink
from glad.opener import URLOpener
from glad.util import makefiledir

if sys.version_info >= (3, 0):
    from urllib.parse import urlencode
else:
    from urllib import urlencode


class NullConfig(Config):
    pass


class BaseGenerator(object):
    DISPLAY_NAME = None
    Config = NullConfig

    def __init__(self, path, opener=None, gen_info_factory=None):
        self.path = os.path.abspath(path)

        self.opener = opener
        if self.opener is None:
            self.opener = URLOpener.default()

        self.gen_info_factory = gen_info_factory or GenerationInfo.create

    @property
    def name(self):
        return self.DISPLAY_NAME

    @property
    def id(self):
        raise NotImplementedError

    def select(self, spec, api, version, profile, extensions, config, sink=LoggingSink(__name__)):
        """
        Basically equivalent to `Specification.select` but gives the generator
        a chance to add additionally required extension, modify the result, etc.

        :param spec: Specification to use
        :param api: API name
        :param version: API version, None means latest
        :param profile: desired profile
        :param extensions: a list of desired extension names, None means all
        :param config: instance of of config specified in `CONFIG`
        :param sink: sink used to collect non fatal errors and information
        :return: FeatureSet with the required types, enums, commands/functions
        """
        return spec.select(api, version, profile, extensions, sink=sink)

    def generate(self, spec, feature_set, config, sink=LoggingSink(__name__)):
        """
        Generates a feature set with the generator.

        :param spec: specification of `feature_set`
        :param feature_set: feature set to generate
        :param config: instance of config specified in `CONFIG`
        :param sink: sink used to collect non fatal errors and information
        """
        raise NotImplementedError


def _api_filter(api):
    if len(api) > 5:
        return api.capitalize()
    return api.upper()


class JinjaGenerator(BaseGenerator):
    TEMPLATES = None

    def __init__(self, path, opener=None, gen_info_factory=None):
        BaseGenerator.__init__(self, path, opener=opener, gen_info_factory=gen_info_factory)

        assert self.TEMPLATES is not None
        self.environment = Environment(
            loader=ChoiceLoader(list(map(PackageLoader, self.TEMPLATES))),
            extensions=['jinja2.ext.do'],
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
            autoescape=False
        )

        self.environment.globals.update(
            set_=set,
            zip=zip
        )

        self.environment.tests.update(
            existsin=lambda value, other: value in other
        )

        self.environment.filters.update(
            api=_api_filter
        )

    @property
    def id(self):
        raise NotImplementedError

    def get_templates(self, spec, feature_set, config):
        """
        Return a list of destination and template tuples for the
        desired feature set and configuration.

        :param spec: specification
        :param feature_set: feature set
        :param config: configuration
        :return: [(destination, name)]
        """
        raise NotImplementedError

    def modify_feature_set(self, spec, feature_set, config):
        """
        Called before `get_templates` and for every `generate` call.
        Mainly useful to update definitions in order to make the
        template interpret types correctly.

        Even though it is possible to return a new feature set,
        such modifications should rather be done in `select`.

        Default implementation does nothing.

        :param feature_set: feature set to modify (the one passed to `get_templates`)
        :return: modified feature set
        """
        return feature_set

    def get_template_arguments(self, spec, feature_set, config):
        return dict(
            spec=spec,
            feature_set=feature_set,
            options=config.to_dict(transform=lambda x: x.lower()),
            gen_info=self.gen_info_factory(self, spec, feature_set, config)
        )

    def generate(self, spec, feature_set, config, sink=LoggingSink(__name__)):
        feature_set = self.modify_feature_set(spec, feature_set, config)
        for template, output_path in self.get_templates(spec, feature_set, config):
            #try:
            template = self.environment.get_template(template)
            #except TemplateNotFound:
            #    # TODO better error, maybe let get_templates throw
            #    raise ValueError('Unsupported specification/configuration')

            result = template.render(
                **self.get_template_arguments(spec, feature_set, config)
            )

            output_path = os.path.join(self.path, output_path)
            makefiledir(output_path)
            with open(output_path, 'w') as f:
                f.write(result)

        self.post_generate(spec, feature_set, config)

    def post_generate(self, spec, feature_set, config):
        pass


class GenerationInfo(object):
    def __init__(self, generator_name, generator_id, spec, info, options, extensions,
                 when=None, commandline=None, online=None):
        """
        Collection of information used to describe a single "generation".
        All held information should either be a string or be "stringifyable".

        :param generator_name: the generator name
        :param generator_id: the generator id (as it was registered with glad.plugin)
        :param spec: the specification name
        :param info: feature set information, usually glad.parse.FeatureSetInfo
        :param options: dictionary containing all enabled options and their value
        :param when: datetime when the code was generated, defaults to now
        :param commandline: callable used to build commandline parameters (will be pased this instance)
        :param online: callable used to build online parameters (will be passed this instance)
        """
        self.generator_name = generator_name
        self.generator_id = generator_id
        self.specification = spec
        self.info = info
        self.options = options
        self.extensions = extensions
        self.when = when or datetime.now().strftime('%c')

        self._commandline = commandline or Commandline()
        self._online = online or Online()

    @classmethod
    def create(cls, generator, spec, feature_set, config, **kwargs):
        return cls(
            generator.name,
            generator.id,
            spec.name,
            feature_set.info,
            config.to_dict(),
            [ext.name for ext in feature_set.extensions],
            **kwargs
        )

    @property
    def version(self):
        return glad.__version__

    @property
    def commandline(self):
        return self._commandline.build(self)

    @property
    def online(self):
        return self._online.build(self)


class ParameterBuilder(object):
    def build(self, gen_info):
        raise NotImplementedError

    def __call__(self, gen_info):
        return self.build(gen_info)


class NullParameterBuilder(ParameterBuilder):
    def build(self, gen_info):
        return ''


class Commandline(ParameterBuilder):
    def __init__(self):
        """
        Parameter builder which serializes a GeneratorInfo
        into commandline arguments.
        """
        pass

    def format_argument(self, name, value):
        name = name.lower().replace('_', '-')

        if isinstance(value, bool):
            return '--{name}'.format(name=name) if value else None

        if isinstance(value, (list, tuple)):
            value = ','.join(str(element) for element in value)

        return '--{name}=\'{value}\''.format(name=name, value=value)

    def build(self, gen_info):
        args = []

        def push(name, value):
            formatted = self.format_argument(name, value)
            if formatted is not None:
                args.append(formatted)

        # general options
        push('merge', gen_info.info.merged)
        push('api', list(gen_info.info))
        push('extensions', gen_info.extensions)

        # generator options
        args.append(gen_info.generator_id)
        for name, value in gen_info.options.items():
            push(name, value)

        return ' '.join(args)


class Online(ParameterBuilder):
    def __init__(self, base_url='http://glad.sh'):
        """
        Parameter builder which serializes a GeneratorInfo
        into commandline arguments.

        :param base_url: base url of the web generator.
        """
        self.base_url = base_url

        self._max_len_threshold = 2000

    def format_argument(self, name, value):
        name = name.lower().replace('-', '_')

        if isinstance(value, bool):
            return name, 'on' if value else 'off'

        if isinstance(value, (list, tuple)):
            result = list()
            for element in value:
                if isinstance(element, (list, tuple)) and len(element) == 2:
                    if isinstance(element[1], bool):
                        if element[1]:
                            result.append(element[0].upper())
                    else:
                        result.append('{0}={1}'.format(*element))
                else:
                    result.append(str(element))
            value = ','.join(result)

        return name, value

    def build(self, gen_info):
        args = collections.OrderedDict()

        def push(name, value):
            name, value = self.format_argument(name, value)
            args[name] = value

        # general options
        push('api', list(gen_info.info))
        push('extensions', gen_info.extensions)

        # generator options
        push('generator', gen_info.generator_id)

        options = [('merge', gen_info.info.merged)]
        options.extend(gen_info.options.items())
        push('options', options)

        def build_url():
            return '{base_url}/#{data}'.format(
                base_url=self.base_url.rstrip('/'),
                data=urlencode(args)
            )

        url = build_url()
        if self._max_len_threshold and len(url) > self._max_len_threshold:
            args.pop('extensions')
            url = build_url()

        return url



from collections import OrderedDict
import jinja2
import re

if hasattr(jinja2, 'pass_context'):
    jinja2_contextfunction = jinja2.pass_context
    jinja2_contextfilter = jinja2.pass_context
else:
    jinja2_contextfunction = jinja2.contextfunction
    jinja2_contextfilter = jinja2.contextfilter


def is_device_command(self):
    """
    Returns true if the command is a Vulkan device command.

    :return: boolean indicating if the command is a device command
    """
    if len(self.params) == 0:
        return False

    first_param = self.params[0]
    # See: https://cgit.freedesktop.org/mesa/mesa/tree/src/intel/vulkan/anv_entrypoints_gen.py#n434
    return first_param.type.type in ('VkDevice', 'VkCommandBuffer', 'VkQueue')


def strip_specification_prefix(name, spec_name=None):
    """
    Used to strip the specification name prefix from a
    command name and extension/feature name.

    E.g.:
        glFoo -> Foo
        GL_ARB_asd -> ARB_asd
        GL_3DFX_tbuffer -> _3DXF_tbffer

    :param name: input name to strip prefix from
    :param spec_name: name of the specification or a Specification
    :return: stripped name
    """
    api_prefix = getattr(spec_name, 'name', spec_name)

    if name.lower().startswith(api_prefix):
        name = name[len(api_prefix):].lstrip('_')

    # 3DFX_tbuffer -> _3DFX_tbuffer
    if not name[0].isalpha():
        name = '_' + name

    return name


def collect_alias_information(commands):
    # Thanks @derhass
    # https://github.com/derhass/glad/commit/9302dc566c695aebece901809f170297627950c9#diff-25f472d6fbc5268fe9a449252923b693

    # keep a dictionary, store the set of aliases known for each function
    # initialize it to identity, each function aliases itself
    alias = dict((command.name, set([command.name])) for command in commands)
    # now, add all further aliases
    for command in commands:
        if command.alias is not None:
            # aliases is the set of all aliases known for this function
            aliases = alias[command.name]
            aliases.add(command.alias)
            # unify all alias sets of all aliased functions
            new_aliases = set()
            missing_funcs = set()
            for aliased_func in aliases:
                try:
                    new_aliases.update(alias[aliased_func])
                except KeyError:
                    missing_funcs.add(aliased_func)
            # remove all missing functions
            new_aliases = new_aliases - missing_funcs
            # add the alias set to all aliased functions
            for aliased_command in new_aliases:
                alias[aliased_command] = new_aliases
    # remove self-aliases and (then) empty entries
    for command in commands:
        if len(alias[command.name]) == 1:
            del alias[command.name]

    return OrderedDict(
        (command.name, sorted(alias[command.name]))
        for command in commands if command.name in alias
    )


def find_extensions_with_aliases(spec, api, version, profile, extensions):
    """
    Finds all extensions that contain a command that is an alias
    to a command in the current feature set (api, version, profile, extensions).

    The resulting list of extensions can be added to the extensions
    list to generate a feature set which contains as many aliases
    as possible.

    :param spec: the specification
    :param api: the requested api
    :param version: the requested version
    :param profile: the requested profile
    :param extensions: the base extension list
    :return: all extensions that contain an alias to the desired feature set
    """
    feature_set = spec.select(api, version, profile, extensions)

    command_names = set(command.name for command in feature_set.commands)

    new_extensions = set()
    for extension in spec.extensions[api].values():
        if extension in feature_set.extensions:
            continue

        for command in extension.get_requirements(spec, api, profile).commands:
            # find all extensions which have an alias to a selected function
            if command.alias and command.alias in command_names:
                new_extensions.add(extension.name)
                break

            # find all extensions that have a function with the same name
            if command.name in command_names:
                new_extensions.add(extension.name)
                break

    return new_extensions




from glad.sink import LoggingSink

try:
    from lxml import etree
    from lxml.etree import ETCompatXMLParser as parser

    def xml_fromstring(argument):
        return etree.fromstring(argument, parser=parser())
    def xml_parse(path):
        return etree.parse(path, parser=parser()).getroot()
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree

    def xml_fromstring(argument):
        return etree.fromstring(argument)
    def xml_parse(path):
        return etree.parse(path).getroot()

import re
import copy
import logging
import os.path
import warnings
from collections import defaultdict, OrderedDict, namedtuple, deque
from contextlib import closing
from itertools import chain

from glad.opener import URLOpener
from glad.util import Version, topological_sort, memoize
import glad.util

logger = logging.getLogger(__name__)


_ARRAY_RE = re.compile(r'\[(\d+)\]')


class FeatureSetInfo(object):
    class InfoItem(namedtuple('InfoItem', ['api', 'version', 'profile', 'identifier'])):
        def __str__(self):
            result = self.api
            if self.profile:
                result += ':{}'.format(self.profile)
            result += '={version.major}.{version.minor}'.format(version=self.version)
            if self.identifier:
                result += '-{!r}'.format(self.identifier)
            return result
        __repr__ = __str__

    def __init__(self, items, merged=False):
        self._items = OrderedDict()
        for item in items:
            self._items.setdefault(item.api, []).append(item)

        self.merged = merged

    @classmethod
    def one(cls, api, version, profile, identifier=None):
        return cls([FeatureSetInfo.InfoItem(api, version, profile, identifier)])

    @property
    def apis(self):
        return list(self._items.keys())

    def __str__(self):
        return repr(list(self))
    __repr__ = __str__

    def __iter__(self):
        return iter(chain.from_iterable(self._items.values()))


class FeatureSet(object):
    def __init__(self, name, info, features, extensions, types, enums, commands):
        self.name = name
        self.info = info
        self.features = features
        self.extensions = extensions
        self.types = types
        self.enums = enums
        self.commands = commands

    def __str__(self):
        return 'FeatureSet(name={self.name}, info={self.info}, extensions={extensions})' \
            .format(self=self, extensions=len(self.extensions))
    __repr__ = __str__

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # good enough for now
        return hash((
            self.info,
            len(self.features), len(self.extensions), len(self.types), len(self.enums), len(self.commands)
        ))

    @property
    @memoize(method=True)
    def _all_enums(self):
        """
        Vulkan introduced grouping of enumerations, they turned from
        basic `#define`'s into actual enumerations `enum Foo { ... }`.

        Glad interprets the "actual" enumerations as types, it's
        an enumeration type with values. But sometimes it is necessary
        to just have all enumerations without grouping information
        available (e.g. for quickly looking up a value). This lives
        under the assumption that enum names are unique.

        :return: a dictionary of name:enum pairs.
        """
        result = dict()
        for enum in self.enums:
            result[enum.name] = enum

        apis = [info.api for info in self.info]
        for type_ in self.types:
            if type_.category == 'enum':
                for enum in type_.enums:
                    # Safety check, this should never happen.
                    # Even though this could happen if you go by the spec,
                    # glad generates duplicate types (and commands) with different APIs
                    # attributes. So the only way this could happen by merging multiple
                    # APIs, but the merging code already makes sure it selects
                    # one of the candidates and not both
                    if enum.name in result:
                        raise ValueError('duplicate enum {} in feature set'.format(enum.name))
                    result[enum.name] = enum

        return result

    def find_enum(self, name, default=None):
        """
        Finds any enum, this includes enums that are part of
        types, by its name.

        :param enum_name: name of the enum
        :param default: default value to return if not found
        :return: the enum if found or else the default
        """
        if name is None:
            return default
        return self._all_enums.get(name, default)

    @classmethod
    def merge(cls, feature_sets, sink=LoggingSink()):
        def to_ordered_dict(items):
            return OrderedDict((item.name, item) for item in items)

        def merge_items(items, new_items):
            for new_item in new_items:
                # TODO merge strategy, prefer khronos types
                in_dict = items.setdefault(new_item.name, new_item)
                if not in_dict is new_item:
                    if not in_dict.is_equivalent(new_item):
                        sink.warning('potential incompatibility: {!r} <-> {!r}'.format(new_item, in_dict))

        feature_set = feature_sets[0]
        others = feature_sets[1:]

        info = list(feature_set.info)
        features = to_ordered_dict(feature_set.features)
        extensions = to_ordered_dict(feature_set.extensions)
        types = to_ordered_dict(feature_set.types)
        enums = to_ordered_dict(feature_set.enums)
        commands = to_ordered_dict(feature_set.commands)

        for other in others:
            if other.info not in info:
                info.extend(other.info)
                merge_items(features, other.features)
                merge_items(extensions, other.extensions)
                merge_items(types, other.types)
                merge_items(enums, other.enums)
                merge_items(commands, other.commands)

        name = os.path.commonprefix(list(chain([feature_set.name], [f.name for f in others])))
        if not name:
            name = feature_set.name

        return FeatureSet(
            name,
            FeatureSetInfo(info, merged=True),
            list(features.values()),
            list(extensions.values()),
            list(types.values()),
            list(enums.values()),
            list(commands.values())
        )


class TypeEnumCommand(namedtuple('_TypeEnumCommand', ['types', 'enums', 'commands'])):
    def __contains__(self, item):
        return item in self.types or item in self.enums or item in self.commands


class Specification(object):
    API = 'https://cvs.khronos.org/svn/repos/ogl/trunk/doc/registry/public/api/'
    NAME = None

    def __init__(self, root):
        self.root = root

        self._combined = None

    def _magic_require(self, api, profile):
        """
        The specifications use a requirement system for most symbols,
        unfortunately this is only partially used for types.

        This method exists to fix the requirement system and
        require the symbols which are not explicitly required
        but yet are needed.

        By default requires all types which are not associated with a
        profile or are associated with the requested profile.

        The problem is some types should *only* be included through
        the requirement system (like khrplatform). If this is the case
        this method should be overwritten.

        :param api: requested api
        :param profile: requested profile
        :return: a requirement or None
        """
        requirements = [name for name, types in self.types.items()
                        if any(t.api in (None, api) for t in types)]

        return Require(api, profile, requirements)

    def _magic_are_enums_blacklisted(self, enums_element):
        """
        Some specifications (Vulkan) have types referring to enums,
        usually by type. To not include these enum types as enum
        (they are already a type), this blacklist exists.

        :return: boolean if enums element is blacklisted
        """
        return False

    @property
    def name(self):
        if self.NAME is None:
            raise NotImplementedError

        return self.NAME

    @classmethod
    def from_url(cls, url, opener=None):
        if opener is None:
            opener = URLOpener.default()

        with closing(opener.urlopen(url)) as f:
            raw = f.read()

        return cls(xml_fromstring(raw))

    @classmethod
    def from_remote(cls, opener=None):
        return cls.from_url(cls.API + cls.NAME + '.xml', opener=opener)

    @classmethod
    def from_string(cls, string):
        return cls(xml_fromstring(string))

    @classmethod
    def from_file(cls, path_or_file_like, opener=None):
        try:
            return cls.from_url('file:' + path_or_file_like, opener=opener)
        except TypeError:
            return cls(xml_parse(path_or_file_like))

    @property
    def comment(self):
        return self.root.find('comment').text

    @property
    def groups(self):
        warnings.warn('the groups have been deprecated in the spec', DeprecationWarning)
        if self._groups is None:
            self._groups = dict((element.attrib['name'], Group(element))
                                for element in self.root.find('groups'))
        return self._groups

    @property
    @memoize(method=True)
    def platforms(self):
        platforms = dict()

        pe = self.root.find('platforms')
        if pe is None:
            pe = []

        for element in pe:
            platform = Platform.from_element(element)
            platforms[platform.name] = platform

        return platforms

    @property
    @memoize(method=True)
    def types(self):
        types = OrderedDict()
        for element in filter(lambda e: e.tag == 'type', iter(self.root.find('types'))):
            name = element.get('name') or element.find('name').text

            if element.get('category') != 'enum':
                types.setdefault(name, list()).extend(Type.from_element(element))
                continue

            # Enum special handling
            enums_element = self.root.findall('.//enums[@type][@name="{}"]'.format(name))
            if len(enums_element) == 0:
                if element.get('alias') is None:
                    # yep the type exists but there is actually no enum for it...
                    logger.debug('type {} with category enum but without <enums>'.format(name))
                    continue
                # It's just a simple alias, add it
                types.setdefault(name, list()).extend(Type.from_element(element))
                continue
            elif len(enums_element) > 1:
                # this should never happen, if it does ... well shit
                raise ValueError('multiple enums with type attribute and name {}'.format(name))
            enums_element = enums_element[0]

            # Copy the type for each API it is used in
            apis = set(e.get('api') for e in self.root.findall('.//require/enum[@extends="{}"]'.format(name)))
            for api in apis or [None]:
                # Hack: Enum Type always returns only one element,
                # because we do the special handling here
                t = EnumType.from_element(element)[0]

                # Assumption for now that enum types do not have an API
                # if there ever is one, the `apis` list needs to be adjusted to only contain t.api
                assert t.api is None
                t.api = api
                t.bitwidth = enums_element.get('bitwidth')

                kwargs = dict(namespace=enums_element.get('namespace'),
                              parent_group=enums_element.get('group'),
                              vendor=enums_element.get('vendor'),
                              comment=enums_element.get('comment', ''))

                enums = OrderedDict()
                for e in (Enum.from_element(e, parent_type=t.name, **kwargs) for e in enums_element.findall('enum')):
                    enums[e.name] = e

                for extension in self.root.findall('.//require/enum[@extends="{}"]/../..'.format(t.name)):
                    try:
                        extnumber = int(extension.attrib['number'])
                    except ValueError:
                        # Most likely a feature, if that happens every extending enum needs
                        # to specify its own extnumber
                        extnumber = None

                    for extending_enum in extension.findall('.//require/enum[@extends="{}"]'.format(t.name)):
                        enum = Enum.from_element(extending_enum, extnumber=extnumber, parent_type=t.name)
                        # Make sure this enum is relevant for the current API
                        if not enum.api == api and enum.api is not None:
                            continue

                        if enum.name not in enums:
                            enums[enum.name] = enum
                        else:
                            # technically not required, but better throw more
                            # than generate broken code because of a broken specification
                            if not enum.value == enums[enum.name].value:
                                raise ValueError('extension enum {} required multiple times '
                                                 'with different values'.format(e.name))

                        enums[enum.name].also_extended_by(extension.attrib['name'])

                t.enums = list(enums.values())

                types.setdefault(t.name, list()).append(t)

        def _type_dependencies(item):
            # all requirements of all types (types can exist more than once, e.g. specialized for an API)
            # but only requirements that are types as well
            requirements = set(r for r in chain.from_iterable(t.requires for t in item[1]) if r in types)
            # requirements.add(r.parent_type)
            aliases = set(t.alias for t in item[1] if t.alias and t.alias in types)
            dependencies = requirements.union(aliases)
            dependencies.discard(item[0])
            return dependencies

        return OrderedDict(topological_sort(types.items(), lambda x: x[0], _type_dependencies))

    @property
    def commands(self):
        commands = dict()
        for element in self.root.find('commands'):
            parsed = Command.from_element(element)
            if len(parsed) > 0:
                commands.setdefault(parsed[0].name, []).extend(parsed)

        # fixup aliases
        for command in chain.from_iterable(commands.values()):
            if command.alias is not None and command.proto is None:
                aliased_command = command
                while aliased_command.proto is None:
                    aliased_command = next(c for c in commands[aliased_command.alias] if c.api == command.api)

                command.proto = Proto(command.name, copy.deepcopy(aliased_command.proto.ret))
                command.params = copy.deepcopy(aliased_command.params)

        return commands

    @property
    @memoize(method=True)
    def enums(self):
        enums = dict()
        for element in self.root.iter('enums'):
            # check if the enum is actually a type
            if self._magic_are_enums_blacklisted(element):
                continue

            namespace = element.get('namespace')
            group = element.get('group')
            vendor = element.get('vendor')
            comment = element.get('comment', '')

            for enum in element:
                if enum.tag in ('unused', 'comment'):
                    continue
                assert enum.tag == 'enum'

                name = enum.attrib['name']
                enums.setdefault(name, []).append(Enum.from_element(
                    enum, namespace=namespace, parent_group=group, vendor=vendor, comment=comment
                ))

        # add enums added through a <require>
        for element in self.root.findall('.//require/enum'):
            # TODO element can be an alias (defined in value) to an extending enum
            if element.get('extends'):
                continue

            enum = Enum.from_element(element)
            enums.setdefault(enum.name, []).append(enum)

        return enums

    @property
    @memoize(method=True)
    def features(self):
        features = defaultdict(dict)
        for element in self.root.iterfind('feature'):
            num = Version(*map(int, element.attrib['number'].split('.')))
            for api in element.attrib['api'].split(','):
                features[api][num] = Feature.from_element(element)

        for api, api_features in features.items():
            features[api] = OrderedDict(sorted(api_features.items(), key=lambda x: x[0]))

        return features

    def highest_version(self, api):
        return sorted(self.features[api].keys(), reverse=True)[0]

    @property
    @memoize(method=True)
    def extensions(self):
        extensions = defaultdict(dict)
        for element in self.root.find('extensions'):
            extension = Extension.from_element(element)
            for api in extension.supported:
                extensions[api][element.attrib['name']] = extension

        return extensions

    def is_extension(self, api, extension_name):
        return extension_name in self.extensions[api]

    def profiles_for_api(self, api):
        profiles = set()

        for feature in chain(self.features[api].values(), self.extensions[api].values()):
            for r in chain(getattr(feature, 'removes', []), feature.requires):
                if (r.api is None or r.api == api) and r.profile is not None:
                    profiles.add(r.profile)

        return profiles

    def protections(self, symbol, api=None, profile=None, feature_set=None):
        """
        Returns a list of protections for a symbol.

        Specifications that do not carry protection info
        may choose to override this function and always
        return an empty list.

        :param symbol: symbol like Enum, Type, Extension etc.
        :param api: api to evaluate for
        :param profile: profile to evaluate for
        :param feature_set: evaluate protections based on feature_set
        :return: a list of protection names
        """
        if getattr(symbol, 'protect', []):
            return symbol.protect

        if getattr(symbol, 'platform', None):
            return [self.platforms[symbol.platform].protect]

        if feature_set:
            extensions = chain(feature_set.features, feature_set.extensions)
        else:
            extensions = chain(self.features, self.extensions)

        protections = list()
        for ext in extensions:
            requirements = ext.get_requirements(self, api=api, profile=profile, feature_set=feature_set)

            if symbol in requirements:
                if ext.protect:
                    protections.extend(ext.protect)
                elif ext.platform is not None:
                    protections.append(self.platforms[ext.platform].protect)
                else:
                    # symbol found in at least one unprotected extension
                    return list()

        return protections

    def find(self, require, api, profile, recursive=False):
        """
        Find all requirements of a require 'instruction'.

        :param require: the require instruction to resolve
        :param api: the api to resolve for
        :param profile: the profile to resolve for
        :param recursive: recursively resolve requirements
        :return: iterator with all results
        """
        if not ((require.profile is None or require.profile == profile) and
                (require.api is None or require.api == api)):
            return

        if self._combined is None:
            self._combined = dict()
            self._combined.update(self.types)
            self._combined.update(self.commands)
            self._combined.update(self.enums)

        requirements = set(require.requirements)
        open_requirements = deque(requirements)
        while open_requirements:
            name = open_requirements.popleft()

            if name in self._combined:
                results = self._combined[name]

                # Get the best match from a list of results, e.g.:
                #  <type name="foo" />
                #  <type name="foo" api="gles" />
                # So here we would go for the latter definition for the API gles.
                best_match = None
                for result in results:
                    # no match so far and result is a match
                    if best_match is None and (result.api is None or result.api == api):
                        best_match = result
                        continue

                    # we had a previous match, is it better?
                    # a result is perfect when the APIs are matching
                    if result.api == api:
                        best_match = result
                        # perfect match -> stop
                        break

                # this should never happen and indicates broken XML!?
                assert best_match is not None

                # maybe we got more requirements (types can require other types)
                # TODO check if _magic_require is still necessary with recursive
                # TODO auto-require types from commands etc.
                requires = getattr(best_match, 'requires', None)
                if recursive and requires is not None:
                    # just add new requirements
                    new_requirements = set(requires).difference(requirements)
                    if new_requirements:
                        requirements.update(new_requirements)
                        open_requirements.extend(new_requirements)

                # Everything that is not a command alias is generated as a define, typedef etc.
                # and not "copy-pasted", this means we have to resolve the alias of a type in order
                # to be able to alias it properly.
                # Commands are simply generated again instead of aliased.
                # The `isinstance` check can be replaced with `if not alias in self.commands`.
                if not isinstance(best_match, Command):
                    alias = getattr(best_match, 'alias', None)
                    if recursive and alias is not None:
                        if alias not in requirements:
                            requirements.add(alias)
                            open_requirements.append(alias)

                yield best_match

    @property
    @memoize(method=True)
    def _all_enums(self):
        """
        Vulkan introduced grouping of enumerations, they turned from
        basic `#define`'s into actual enumerations `enum Foo { ... }`.

        Glad interprets the "actual" enumerations as types, it's
        an enumeration type with values. But sometimes it is necessary
        to just have all enumerations without grouping information
        available (e.g. for quickly looking up a value). This lives
        under the assumption that enum names are unique.

        :return: a dictionary of name:enum pairs.
        """
        result = dict()

        for type_ in self.types.values():
            type_ = type_[0]
            if type_.category == 'enum':
                for enum in type_.enums:
                    result[enum.name] = enum

        return result

    def find_enum(self, name, default=None):
        """
        Finds any enum, this includes enums that are part of
        types, by its name.

        :param name: name of the enum
        :param default: default value to return if not found
        :return: the enum if found or else the default
        """
        if name is None:
            return default
        return self._all_enums.get(name, default)

    @staticmethod
    def split_types(iterable, types):
        result = tuple(set() for _ in types)

        for obj in iterable:
            for i, type_ in enumerate(types):
                if isinstance(obj, type_):
                    result[i].add(obj)

        return result

    def select(self, api, version, profile, extension_names, sink=LoggingSink(__name__)):
        """
        Select a specific configuration from the specification.

        :param api: API name
        :param version: API version, None means latest
        :param profile: desired profile
        :param extension_names: a list of desired extension names, None means all
        :param sink: sink used to store informations, warnings and errors that are not fatal
        :return: FeatureSet with the required types, enums, commands/functions
        """
        # make sure that there is a profile if one is required/available
        available_profiles = self.profiles_for_api(api)
        if len(available_profiles) == 1 and profile is None:
            # If there is only one profile, use that
            profile = next(iter(available_profiles))
        if available_profiles and profile not in available_profiles:
            if profile is None:
                raise ValueError(
                    'Profile required for {!r}-API, one of {!r}'
                    .format(api, tuple(available_profiles))
                )
            raise ValueError(
                'Invalid profile {!r} for {!r}-API, expected one of {!r}'
                .format(profile, api, tuple(available_profiles))
            )

        if not self.features.get(api):
            raise ValueError('Invalid API {!r} for specification {!r}'.format(api, self.name))

        # None means latest version, update the dictionary with the latest version
        if version is None:
            version = self.highest_version(api)
            sink.info('no explicit version given for api {}, using {}'.format(api, version))

        # make sure the version is valid
        if version not in self.features[api]:
            raise ValueError(
                'Unknown version {}.{} for API {} (specification: {})'
                .format(version[0], version[1], api, self.name)
            )

        all_extensions = list(self.extensions[api].keys())
        if extension_names is None:
            sink.info('adding all extensions for api {} to result'.format(api))
            # None means all extensions
            extension_names = all_extensions
        else:
            # make sure only valid extensions are listed
            for extension in extension_names:
                if extension not in all_extensions:
                    raise ValueError(
                        'Invalid extension {!r} for specification {!r}'.format(
                            extension, self.name
                        )
                    )

        # remove duplicates
        extension_names = set(extension_names)

        # Remove weird GLX extensions which use undefined types
        for extension in ['GLX_SGIX_video_source', 'GLX_SGIX_dmbuffer']:
            try:
                extension_names.remove(extension)
            except KeyError:
                pass
            else:
                sink.warning('removed extension {}, it uses unsupported types'.format(extension))

        # OpenGL version 3.3 includes all versions up to 3.3
        # Collect a list of all required features grouped by API
        features = [feature for fversion, feature in self.features[api].items()
                    if fversion <= version]

        # Collect a list of extensions grouped by API
        extensions = [self.extensions[api][name] for name in extension_names
                      if name in self.extensions[api]]

        # Collect information
        result = set()
        # collect all required types, functions (=commands) and enums by API
        # features are special extensions
        for extension in chain(features, extensions):
            for require in extension.requires:
                found = self.find(require, api, profile, recursive=True)
                result = result.union(found)

            for remove in getattr(extension, 'removes', []):
                if ((remove.profile is None or remove.profile == profile) and
                        (remove.api is None or remove.api == api)):
                    result = result.difference(remove.removes)

        # At this point one could hope that the XML files would be sane, but of course they are not!?
        # There is a builtin requirement system which is used for functions and enums,
        # but only partially for types WHY!??!?!?!??!?!
        # So let's fix this here ....
        # require all types which are not associated with a profile or with this one specific
        # and to make it more fun, there are a few types which should only be included through
        # the requirement system (less includes .. mainly important for khrplatform)
        require = self._magic_require(api, profile)
        if require is not None:
            # find and add the requirements just defined
            result = result.union(self.find(require, api, profile, recursive=True))

        # OH WAIT THERE IS MORE!? E.g. Opengl 1.0 HAS *ZERO* Enums? Why?
        # I dont know, maybe some lazy ass who didnt want to figure out which enums were introduced
        # in Opengl 1.1 and just added all of them to 1.1 and none to 1.0
        # We do nothing here and hope 1.0 will stay an exception ...

        # Split information into types, enums and commands/function
        types, enums, commands = self.split_types(
            result, types=(Type, Enum, Command)
        )

        # We need to sort the types since a few definitions depend on other types
        all_sorted_types = list(self.types.keys())
        types = sorted(types, key=all_sorted_types.index)

        features = sorted(features, key=lambda x: x.name)
        extensions = sorted(extensions, key=lambda x: x.name)
        enums = sorted(enums, key=lambda x: x.name)
        commands = sorted(commands, key=lambda x: x.name)

        return FeatureSet(api, FeatureSetInfo.one(api, version, profile),
                          features, extensions, types, enums, commands)


class Group(object):
    def __init__(self, element):
        self.name = element.attrib['name']
        self.enums = [enum.attrib['name'] for enum in element]


# required for set operations in select (union/difference)
# TODO find out if this is problematic
class IdentifiedByName(object):
    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        name = getattr(other, 'name', other)
        return self.name == name


class Platform(object):
    def __init__(self, name, protect, comment=''):
        self.name = name
        self.protect = protect
        self.comment = comment

    @classmethod
    def from_element(cls, element):
        name = element.attrib['name']
        protect = element.attrib['protect']
        comment = element.get('comment', '')
        return Platform(name, protect, comment=comment)


class Type(IdentifiedByName):
    _FACTORIES = dict()

    @staticmethod
    def register(category, type_factory):
        Type._FACTORIES[category] = type_factory

    def __init__(self, name, api=None, category=None, alias=None, requires=None, parent=None, apientry=False, raw=None):
        self.name = name
        self.api = api
        self.category = category

        self.alias = alias
        self.requires = requires or []
        self.parent = parent
        self.apientry = apientry

        self._raw = raw

    @classmethod
    def factory(cls, element, name, data):
        return [cls(name, **data)]

    @staticmethod
    def from_element(element):
        apientry = element.find('apientry')
        if apientry is not None:
            # not so great workaround to get APIENTRY included in the raw output
            apientry.text = 'APIENTRY'

        raw = ''.join(element.itertext())
        api = element.get('api')
        category = element.get('category')
        name = element.get('name') or element.find('name').text

        alias = element.get('alias')
        parent = element.get('parent')

        # a type referenced by a struct/funcptr is required by this type
        requires = set(t.text for t in element.iter('type') if t is not element)
        requires.update(e.text for e in element.iter('enum'))
        if element.get('requires'):
            requires.add(element.get('requires'))

        data = dict(
            api=api,
            category=category,
            alias=alias,
            requires=requires,
            parent=parent,
            apientry=apientry is not None,
            raw=raw
        )

        factory = Type._FACTORIES.get(category, Type.factory)
        return factory(element, name, data)

    def is_equivalent(self, other):
        return self._raw == other._raw

    def is_descendant(self, basetype, typeslist):
        if self.name == basetype:
            return True

        cur = self
        while cur.parent is not None:
            if cur.parent == basetype:
                return True
            cur = typeslist[cur.parent][0]
        return False

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Type(raw={self._raw!r})'.format(self=self)


class FuncPointerType(Type):
    _Parameters = namedtuple('Parameters', ['type', 'name'])

    def __init__(self, name, ret=None, parameters=None, **kwargs):
        Type.__init__(self, name, **kwargs)

        self.ret = ret
        self.parameters = parameters or []

    @classmethod
    def factory(cls, element, name, data):
        raw = data['raw'].strip().replace('\r', '').replace('\n', '')

        # typedef {RETURN} (VKAPI_PTR *{NAME})({PARAMS...});

        # extract {RETURN}, split the typedef away,
        # then take everything up to the first (
        ret = raw.split(None, 1)[1].split('(', 1)[0].strip()
        # split to the 2nd (, take everything after,
        # then split to the next ). Leaving {PARAMS...} behind
        parameters = [p for p in raw.split('(', 2)[2].split(')')[0].strip().split(',') if p.strip()]

        # when {PARAMS...} is void
        if len(parameters) == 1 and parameters[0] == 'void':
            parameters = []

        for i, parameter in enumerate(parameters):
            type_, pname = parameter.rsplit(None, 1)
            parameters[i] = FuncPointerType._Parameters(
                type_.replace(' ', ''), pname.strip()
            )

        return [cls(name, ret=ret, parameters=parameters, **data)]


class MemberType(Type):
    def __init__(self, name, members=None, **kwargs):
        Type.__init__(self, name, **kwargs)

        self.members = members or []

    @classmethod
    def factory(cls, element, name, data):
        members = [Member.from_element(e) for e in element.findall('member')]

        # May not have members at all (struct with only an alias)
        if len(members) == 0:
            return [cls(name, **data)]

        result = list()
        for api in set(member.api for member in members):
            api_members = [member for member in members if member.api is None or member.api == api]
            t = cls(name, members=api_members, **data)
            t.api = api
            result.append(t)
        return result


class UnionType(MemberType):
    pass


class StructType(MemberType):
    pass


class EnumType(Type):
    def __init__(self, name, enums=None, bitwidth=None, **kwargs):
        Type.__init__(self, name, **kwargs)

        self.enums = enums or []
        self.bitwidth = bitwidth

    @property
    def expanded_name(self):
        return glad.util.expand_type_name(self.name)

    @memoize(method=True)
    def enums_for(self, feature_set):
        relevant = set(feature_set.features) | set(feature_set.extensions)

        required_names = set()
        result = list()
        # assume order, an alias must be after the enum it is aliased to
        # reverse here so we add the alias first then we can check if the
        # value that is referenced needs to be added as well
        for enum in reversed(self.enums):
            if len(enum.extended_by) == 0 or enum.extended_by & relevant:
                result.insert(0, enum) # restore order
                required_names.add(enum.alias)
            elif enum.name in required_names:
                result.insert(0, enum) # restore order

        return result


class TypedType(Type):
    def __init__(self, name, type=None, **kwargs):
        Type.__init__(self, name, **kwargs)

        self.type = type

    @classmethod
    def factory(cls, element, name, data):
        # may be none if the handle is just aliased
        type_element = element.find('type')
        type_ = type_element.text if type_element is not None else None
        return [cls(name, type=type_, **data)]


class HandleType(TypedType):
    pass


class BasetypeType(TypedType):
    pass


class BitmaskType(TypedType):
    pass


Type.register('funcpointer', FuncPointerType.factory)
Type.register('union', UnionType.factory)
Type.register('struct', StructType.factory)
Type.register('enum', EnumType.factory)
Type.register('handle', HandleType.factory)
Type.register('basetype', BasetypeType.factory)
Type.register('bitmask', BitmaskType.factory)


class Member(IdentifiedByName):
    def __init__(self, name, type_, api=None, enum=None):
        self.name = name
        self.type = type_
        self.api = api
        self.enum = enum

    @classmethod
    def from_element(cls, element):
        type_ = ParsedType.from_element(element)
        enum = element.find('enum')
        api = element.get('api')

        return Member(type_.name, type_, api=api, enum=enum.text if enum is not None else None)

    def __str__(self):
        return 'Member(name={self.name}, type={self.type})'.format(self=self)
    __repr__ = __str__


class Enum(IdentifiedByName):
    BASE_EXTENSION_OFFSET = 1000000000
    EXTENSION_NUMBER_MULTIPLIER = 1000
    EXTENSION_NUMBER_OFFSET = -1

    def __init__(self, name, value, bitpos, api, type_,
                 alias=None, namespace=None, group=None, parent_group=None,
                 vendor=None, comment='', parent_type=None, extended_by=None):
        """
        :param name: name of the enum
        :param value: value of the enum
        :param bitpos: alternative way of specifying the value
        :param api: api as specified on the enum
        :param type_: type of the enum as specified on the element
        :param alias: alias of the enum
        :param namespace: namespace of the group e.g. GL
        :param group: group specified in on the enum, comma separated for multiple
        :param parent_group: if the enum was defined in an <enums> group
        :param vendor: vendor specified on <enums> group
        :param comment: comment specified on the <enums> group
        :param parent_type: parent type if the enums is grouped and not just a global define (Foo.BAR)
        :param extended_by: list of enums this is extended by
        """
        self.name = name
        self.value = value
        if self.value is None and bitpos is not None:
            self.value = str(1 << int(bitpos))
        self.bitpos = bitpos
        self.api = api
        self.type = type_

        self.alias = alias

        self.namespace = namespace
        self.group = group
        self.parent_group = parent_group
        self.vendor = vendor
        self.comment = comment

        # name of the type this enum is a part of
        self.parent_type = parent_type

        self.extended_by = set(extended_by) if extended_by else set()

    @property
    def expanded_name(self):
        return glad.util.expand_type_name(self.name)

    def also_extended_by(self, name):
        self.extended_by.add(name)

    def is_equivalent(self, other):
        return self.name == other.name and self.value == other.value

    @property
    def groups(self):
        """
        Returns a list of parsed groups, group is a comma separated value
        as used in the XML.

        :return: empty list or list of groups
        """
        return [] if self.group is None else self.group.split(',')

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Enum(name={self.name}, value={self.value}, type={self.type})'.format(self=self)

    @classmethod
    def from_element(cls, element, extnumber=None, **kwargs):
        name = element.attrib['name']
        value = element.get('value')
        bitpos = element.get('bitpos')
        api = element.get('api')
        type_ = element.get('type')
        group = element.get('group')

        alias = element.get('alias')

        if element.get('extnumber'):
            extnumber = int(element.get('extnumber'))

        if element.get('offset') is not None:
            if extnumber is None:
                raise ValueError('enum has offset, extnumber is required')

            extbase = (cls.EXTENSION_NUMBER_MULTIPLIER * (extnumber + cls.EXTENSION_NUMBER_OFFSET))
            value = cls.BASE_EXTENSION_OFFSET + extbase + int(element.get('offset'))

            if element.get('dir') == '-':
                value = -value

        if value is not None:
            value = str(value)

        return cls(name, value, bitpos, api, type_, alias=alias, group=group, **kwargs)


class Command(IdentifiedByName):
    def __init__(self, name, api=None, proto=None, params=None, alias=None):
        self.name = name
        self.api = api
        self.proto = proto
        self.params = params
        self.alias = alias

        if self.alias is None and self.proto is None:
            raise ValueError("command is neither a full command nor an alias")

    @classmethod
    def from_element(cls, element):
        params = None
        proto = element.find('proto')
        if proto is not None:
            proto = Proto.from_element(proto)
            params = [Param(ele) for ele in filter(lambda e: e.tag == 'param', iter(element))]

        alias = element.get('alias')
        if alias is None:
            alias_element = element.find('alias')
            if alias_element is not None:
                alias = alias_element.attrib['name']

        api = element.get('api')

        name = element.get('name') or proto.name

        # Only alias or no members
        if params is None or len(params) == 0:
            return [cls(name, api=api, proto=proto, params=params, alias=alias)]

        result = list()
        apis = set(param.api for param in params)
        for api in apis:
            api_params = [param for param in params if param.api is None or param.api == api]
            result.append(cls(name, api=api, proto=proto, params=api_params, alias=alias))

        return result

    @property
    def requires(self):
        if self.params is None:
            return list()

        requires = [param.type.original_type for param in self.params if param.type.original_type]
        if self.proto.ret.original_type:
            requires.append(self.proto.ret.original_type)
        return requires

    def is_equivalent(self, other):
        return self.proto == other.proto and self.params == other.params

    def __str__(self):
        return '{self.proto.name}({params})'.format(
            self=self,
            params=','.join(map(str, self.params))
        )
    __repr__ = __str__


class Proto(object):
    def __init__(self, name, ret):
        self.name = name
        self.ret = ret

    @classmethod
    def from_element(cls, element):
        return Proto(element.find('name').text, ParsedType.from_element(element))

    def is_equivalent(self, other):
        return self.ret == other.ret

    def __str__(self):
        return '{self.ret} {self.name}'.format(self=self)


class Param(object):
    def __init__(self, element):
        self.group = element.get('group')
        self.type = ParsedType.from_element(element)
        self.name = element.find('name').text.strip('*')
        self.api = element.get('api')

    def is_equivalent(self, other):
        return self.type == other.type

    def __str__(self):
        return '{0!r} {1}'.format(self.type, self.name)


class ParsedType(object):
    def __init__(self, name, type_, original_type,
                 is_pointer=0, is_array=0, is_struct=False, is_const=False,
                 is_unsigned=False, comment='', raw=None, element=None):
        self.name = name
        self.original_type = original_type
        self.type = type_

        self.is_pointer = is_pointer
        self.is_array = is_array
        self.is_struct = is_struct
        self.is_const = is_const
        self.is_unsigned = is_unsigned

        self.comment = comment

        self._raw = raw
        self._element = element

    def is_equivalent(self, other):
        return self._raw == other._raw

    @classmethod
    def from_string(cls, raw):
        type_ = raw.replace('const', '') \
            .replace('unsigned', '') \
            .replace('struct', '') \
            .replace('*', '') \
            .strip().split(None, 1)[0]

        # 0 if no pointer, 1 if *, 2 if **
        is_pointer = 0 if raw is None else raw.count('*')
        array_match = _ARRAY_RE.search(raw)
        is_array = 0 if array_match is None else int(array_match.group(1))
        is_const = False if raw is None else 'const' in raw
        is_unsigned = False if raw is None else 'unsigned' in raw
        is_struct = 'struct' in raw

        return cls(None, type_, raw,
                   is_pointer=is_pointer, is_struct=is_struct,
                   is_const=is_const, is_unsigned=is_unsigned,
                   raw=raw)

    @classmethod
    def from_element(cls, element):
        # assume just one comment element
        comment = ' '.join(c.text for c in element.iter('comment'))

        raw = ' '.join(glad.util.itertext(element, ignore=('comment',)))
        name = element.find('name').text

        original_type = None if element.find('type') is None else element.find('type').text

        ptype = element.find('ptype')
        if ptype is not None:
            ptype = ptype.text if ptype is not None else None
            type_ = ptype.replace('struct', '')
        else:
            type_ = raw.replace('const', '') \
                       .replace('unsigned', '') \
                       .replace('struct', '') \
                       .replace('*', '') \
                       .strip().split(None, 1)[0]

        # 0 if no pointer, 1 if *, 2 if **
        is_pointer = 0 if raw is None else raw.count('*')
        array_match = _ARRAY_RE.search(raw)
        is_array = 0 if array_match is None else int(array_match.group(1))
        is_const = False if raw is None else 'const' in raw
        is_unsigned = False if raw is None else 'unsigned' in raw
        is_struct = 'struct' in raw

        return cls(name, type_, original_type,
                   is_pointer=is_pointer, is_array=is_array,
                   is_struct=is_struct, is_const=is_const, is_unsigned=is_unsigned,
                   comment=comment, raw=raw, element=element)


# TODO unify API
class Require(object):
    def __init__(self, api, profile, requirements, extension=None, feature=None, comment=''):
        self.api = api
        self.profile = profile
        self.requirements = requirements

        self.extension = extension
        self.feature = feature

        self.comment = comment

    def is_equivalent(self, other):
        return self.requirements == other.requirements

    @classmethod
    def from_element(cls, element):
        requirements = [child.get('name') for child in element if not child.get('extends')]
        return cls(element.get('api'), element.get('profile'), requirements,
                   element.get('extension'), element.get('feature'), element.get('comment'))


class Remove(object):
    def __init__(self, element):
        self.api = element.get('api')
        self.profile = element.get('profile')

        self.removes = [child.get('name') for child in element]

    def is_equivalent(self, other):
        return self.removes == other.removes


class Extension(IdentifiedByName):
    def __init__(self, name, supported=None, requires=None,
                 type_=None, protect=None, platform=None):
        self.name = name
        self.supported = supported
        self.requires = requires or []
        self.type = type_
        self.protect = protect or []
        self.platform = platform

    @classmethod
    def from_element(cls, element):
        name = element.attrib['name']
        # Vulkan uses `,` and GL uses `|`
        supported = element.attrib['supported'].replace('|', ',').split(',')

        requires = [Require.from_element(require) for require in element.findall('require')]

        type_ = element.get('type')
        protect = [p.strip() for p in element.get('protect', '').split(',') if p.strip()]
        platform = element.get('platform')

        return cls(name, supported=supported, requires=requires,
                   type_=type_, protect=protect, platform=platform)

    def supports(self, api):
        return api in self.supported

    def is_equivalent(self, other):
        return self.requires == other.requires

    @memoize(method=True)
    def get_requirements(self, spec, api=None, profile=None, feature_set=None):
        """
        Find all types, enums and commands/functions which are required
        for this extension/feature.

        :param spec: the specification
        :param api: requested api, takes precedence over feature_set.info
        :param profile: requested profile (requires `api`)
        :param feature_set: used to provide api and profile, if supplied uses
                            all APIs and profiles stored in `feature_set.info`.
                            Also limits the return values to symbols
                            included in the feature set.
        :return TypeEnumCommand: returns a :py:class:`TypeEnumCommand`
        """
        result = set()

        if api is None and feature_set is None:
            raise ValueError('either API or feature_set required')

        for require in self.requires:
            if api is not None:
                result.update(spec.find(require, api, profile, recursive=True))
            else:
                for info in feature_set.info:
                    result.update(spec.find(require, info.api, info.profile, recursive=True))

        types, enums, commands = spec.split_types(result, (Type, Enum, Command))

        if feature_set is None:
            return TypeEnumCommand(
                sorted(types, key=lambda x: x.name),
                sorted(enums, key=lambda x: x.name),
                sorted(commands, key=lambda x: x.name),
            )

        return TypeEnumCommand(
            sorted(types.intersection(feature_set.types), key=lambda x: x.name),
            sorted(enums.intersection(feature_set.enums), key=lambda x: x.name),
            sorted(commands.intersection(feature_set.commands), key=lambda x: x.name),
        )

    def __str__(self):
        return self.name
    __repr__ = __str__


class Feature(Extension):
    def __init__(self, name, api, version, removes=None, **kwargs):
        Extension.__init__(self, name, **kwargs)

        self.api = api
        self.version = version

        self.removes = removes or []

    def is_equivalent(self, other):
        return Extension.is_equivalent(self, other) and self.removes == other.removes

    @classmethod
    def from_element(cls, element, api=None):
        name = element.attrib['name']
        api = set(element.attrib['api'].split(','))
        version = Version(*tuple(map(int, element.attrib['number'].split('.'))))

        requires = [Require.from_element(require) for require in element.findall('require')]
        removes = [Remove(remove) for remove in element.findall('remove')]

        type_ = element.get('type')
        protect = [p.strip() for p in element.get('protect', '').split(',') if p.strip()]
        platform = element.get('platform')

        return cls(name, api, version, supported=api, requires=requires,
                   removes=removes, type_=type_, protect=protect, platform=platform)

    def __str__(self):
        return '{self.name}@{self.version!r}'.format(self=self)
    __repr__ = __str__



import copy
import itertools

import os.path
import re
from collections import namedtuple
from contextlib import closing

from glad.config import Config, ConfigOption, UnsupportedConstraint
from glad.sink import LoggingSink
from glad.generator import JinjaGenerator
from glad.generator.util import (
    is_device_command,
    strip_specification_prefix,
    collect_alias_information,
    find_extensions_with_aliases,
    jinja2_contextfunction,
    jinja2_contextfilter
)
from glad.parse import Type, EnumType
from glad.specification import VK, GL, WGL
import glad.util

_ARRAY_RE = re.compile(r'\[[\d\w]*\]')

DebugArguments = namedtuple('_DebugParams', ['impl', 'function', 'pre_callback', 'post_callback', 'ret'])
DebugReturn = namedtuple('_DebugReturn', ['declaration', 'assignment', 'ret'])


class Header(object):
    def __init__(self, name, include, url, requires=None):
        self.name = name
        self.include = include
        self.url = url
        self.requires = requires


def type_to_c(parsed_type):
    result = ''

    for text in glad.util.itertext(parsed_type._element, ignore=('comment',)):
        if text == parsed_type.name:
            # yup * is sometimes part of the name
            result += '*' * text.count('*')
        else:
            result += text
    result = _ARRAY_RE.sub('*', result)
    return result.strip()


def params_to_c(params):
    result = ', '.join(param.type._raw for param in params) if params else 'void'
    result = ' '.join(result.split())
    return result


def param_names(params):
    return ', '.join(param.name for param in params)


@jinja2_contextfunction
def loadable(context, extensions=None, api=None):
    spec = context['spec']
    feature_set = context['feature_set']

    if extensions is None:
        extensions = (feature_set.features, feature_set.extensions)
    elif len(extensions) > 0:
        # allow loadable(feature_set.features), nicer syntax in templates
        try:
            iter(extensions[0])
        except TypeError:
            extensions = [extensions]

    for extension in itertools.chain.from_iterable(extensions):
        if api is None or extension.supports(api):
            commands = extension.get_requirements(spec, feature_set=feature_set).commands
            if commands:
                yield extension, commands


def is_void(t):
    # lower because of win API having VOID
    return type_to_c(t).lower() == 'void'


def get_debug_impl(command, command_code_name=None):
    command_code_name = command_code_name or command.name

    impl = params_to_c(command.params)
    func = param_names(command.params)

    pre_callback = ', '.join(filter(None, [
        '"{}"'.format(command.name),
        '(GLADapiproc) {}'.format(command_code_name),
        str(len(command.params)),
        func
    ]))

    is_void_ret = is_void(command.proto.ret)

    post_callback = ('NULL, ' if is_void_ret else '(void*) &ret, ') + pre_callback

    ret = DebugReturn('', '', '')
    if not is_void_ret:
        ret = DebugReturn(
            '{} ret;\n    '.format(type_to_c(command.proto.ret)),
            'ret = ',
            'return ret;'
        )

    return DebugArguments(impl, func, pre_callback, post_callback, ret)


@jinja2_contextfilter
def ctx(jinja_context, name, context='context', raw=False, name_only=False, member=False):
    options = jinja_context['options']

    prefix = 'glad_'
    if options['mx']:
        prefix = context + '->'
        if name.startswith('GLAD_'):
            name = name[5:]

        if not raw:
            name = strip_specification_prefix(name, jinja_context['spec'])

    # it's a mx struct member
    if member:
        return name

    # you won't the name, only when we're not mx
    if name_only and not options['mx']:
        return name

    return prefix + name


@jinja2_contextfilter
def pfn(context, value):
    spec = context['spec']
    if spec.name in (VK.NAME,):
        return 'PFN_' + value
    return 'PFN' + value.upper() + 'PROC'


@jinja2_contextfilter
def c_commands(context, commands):
    """
    The c in c_commands refers to the c file.

    This function filters a list of commands for the generated .c file.
    WGL core functions are not dynamically loaded but need to be linked,
    this functions filters out wgl core functions for the .c file.

    :param context: jinja context
    :param commands: list of commands
    :return: commands filtered
    """
    spec = context['spec']
    if not spec.name == WGL.NAME:
        return commands

    feature_set = context['feature_set']
    core = feature_set.features[0].get_requirements(spec, feature_set=feature_set)

    return [command for command in commands if not command in core]


@jinja2_contextfunction
def enum_member(context, type_, member, require_value=False):
    if member.alias is None:
        return member.value

    feature_set = context['feature_set']
    enums_of_type = type_.enums_for(feature_set)

    def is_enum_before(target, before):
        for enum in enums_of_type:
            if enum.name == target:
                return True
            if enum.name == before:
                return False

    if not require_value:
        if is_enum_before(member.alias, member.name):
            return member.alias

    # This is the part where the spec is annoying again
    # an enum that has been moved into core in a later version
    # loses its _KHR postfix, but in an earlier version this still requires an extension...
    # Luckily glad automatically adds the necessary enum to the feature set,
    # but it doesn't get generated, because it is not actually part of the selected feature set.
    # Just have to get the actual value now
    def resolve(target):
        target = feature_set.find_enum(target)
        if target.alias is None:
            return target.value
        return resolve(target.alias)

    return resolve(member.alias)


_CPP_STYLE_COMMENT_RE = re.compile(r'(^|\s|\))//(?P<comment>[^\r^\n]*)', flags=re.MULTILINE)


def replace_cpp_style_comments(inp):
    return _CPP_STYLE_COMMENT_RE.sub(r'\1/*\2 */', inp)


class CConfig(Config):
    DEBUG = ConfigOption(
        converter=bool,
        default=False,
        description='Enables generation of a debug build'
    )
    ALIAS = ConfigOption(
        converter=bool,
        default=False,
        description='Enables function pointer aliasing'
    )
    MX = ConfigOption(
        converter=bool,
        default=False,
        description='Enables support for multiple GL contexts'
    )
    # MX_GLOBAL = ConfigOption(
    #    converter=bool,
    #    default=False,
    #    description='Mimic global GL functions with context switching'
    # )
    HEADER_ONLY = ConfigOption(
        converter=bool,
        default=False,
        description='Generate a header only version of glad'
    )
    LOADER = ConfigOption(
        converter=bool,
        default=False,
        description='Include internal loaders for APIs'
    )
    ON_DEMAND = ConfigOption(
        converter=bool,
        default=False,
        description='On-demand function pointer loading, initialize on use (experimental)'
    )

    __constraints__ = [
        # RequirementConstraint(['MX_GLOBAL'], 'MX'),
        UnsupportedConstraint(['MX'], 'DEBUG'),
        # RequirementConstraint(['MX', 'DEBUG'], 'MX_GLOBAL')
        UnsupportedConstraint(['MX'], 'ON_DEMAND')
    ]


class CGenerator(JinjaGenerator):
    DISPLAY_NAME = 'C/C++'

    TEMPLATES = ['glad.generator.c']
    Config = CConfig

    ADDITIONAL_HEADERS = [
        Header(
            'khrplatform',
            'KHR/khrplatform.h',
            'https://raw.githubusercontent.com/KhronosGroup/EGL-Registry/main/api/KHR/khrplatform.h'
        ),
        Header(
            'eglplatform',
            'EGL/eglplatform.h',
            'https://raw.githubusercontent.com/KhronosGroup/EGL-Registry/main/api/EGL/eglplatform.h'
        ),
        Header(
            'vk_platform',
            'vk_platform.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vulkan/vk_platform.h'
        ),
        Header(
            'vk_video/vulkan_video_codecs_common.h',
            'vk_video/vulkan_video_codecs_common.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codecs_common.h'
        ),
        Header(
            'vk_video/vulkan_video_codec_h264std.h',
            'vk_video/vulkan_video_codec_h264std.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h264std.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_h264std_decode.h',
            'vk_video/vulkan_video_codec_h264std_decode.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h264std_decode.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_h264std_encode.h',
            'vk_video/vulkan_video_codec_h264std_encode.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h264std_encode.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_h265std.h',
            'vk_video/vulkan_video_codec_h265std.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h265std.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_h265std_decode.h',
            'vk_video/vulkan_video_codec_h265std_decode.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h265std_decode.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_h265std_encode.h',
            'vk_video/vulkan_video_codec_h265std_encode.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_h265std_encode.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_av1std.h',
            'vk_video/vulkan_video_codec_av1std.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_av1std.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
        Header(
            'vk_video/vulkan_video_codec_av1std_decode.h',
            'vk_video/vulkan_video_codec_av1std_decode.h',
            'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Headers/main/include/vk_video/vulkan_video_codec_av1std_decode.h',
            requires=['vk_video/vulkan_video_codecs_common.h']
        ),
    ]

    def __init__(self, *args, **kwargs):
        JinjaGenerator.__init__(self, *args, **kwargs)

        self._headers = dict()

        self.environment.globals.update(
            get_debug_impl=get_debug_impl,
            loadable=loadable,
            enum_member=enum_member,
            chain=itertools.chain
        )

        self.environment.filters.update(
            defined=lambda x: 'defined({})'.format(x),
            type_to_c=type_to_c,
            params_to_c=params_to_c,
            param_names=param_names,
            pfn=pfn,
            ctx=ctx,
            no_prefix=jinja2_contextfilter(lambda ctx, value: strip_specification_prefix(value, ctx['spec'])),
            c_commands=c_commands
        )

        self.environment.tests.update(
            supports=lambda x, arg: x.supports(arg),
            void=is_void,
        )

    @property
    def id(self):
        return 'c'

    def select(self, spec, api, version, profile, extensions, config, sink=LoggingSink(__name__)):
        if extensions is not None:
            extensions = set(extensions)

            if api == 'wgl':
                # See issue #40: https://github.com/Dav1dde/glad/issues/40
                # > Currently if you generate a loader without these extensions
                # > (WGL_ARB_extensions_string and WGL_EXT_extensions_string) it won't compile.
                # Adds these 2 extensions if they are missing.
                extensions.update(('WGL_ARB_extensions_string', 'WGL_EXT_extensions_string'))

            if config['ALIAS']:
                extensions.update(find_extensions_with_aliases(spec, api, version, profile, extensions))

        return JinjaGenerator.select(self, spec, api, version, profile, extensions, config, sink=sink)

    def get_template_arguments(self, spec, feature_set, config):
        args = JinjaGenerator.get_template_arguments(self, spec, feature_set, config)

        # TODO allow MX for every specification/api
        if spec.name not in (VK.NAME, GL.NAME):
            args['options']['mx'] = False
            args['options']['mx_global'] = False

        args.update(
            aliases=collect_alias_information(feature_set.commands),
            # required for vulkan loader:
            device_commands=list(filter(is_device_command, feature_set.commands))
        )

        return args

    def get_templates(self, spec, feature_set, config):
        header = 'include/glad/{}.h'.format(feature_set.name)
        source = 'src/{}.c'.format(feature_set.name)

        templates = list()

        if config['HEADER_ONLY']:
            templates.extend([
                ('header_only.h', header)
            ])
        else:
            templates.extend([
                ('{}.h'.format(spec.name), header),
                ('{}.c'.format(spec.name), source)
            ])

        return templates

    def post_generate(self, spec, feature_set, config):
        self._add_additional_headers(feature_set, config)

    def modify_feature_set(self, spec, feature_set, config):
        # TODO this takes rather long (~30%), maybe drop it?
        feature_set = copy.deepcopy(feature_set)

        self._fix_issue_70(feature_set)
        self._fix_cpp_style_comments(feature_set)
        self._fixup_enums(feature_set)
        self._add_header_requirements(feature_set)
        self._replace_included_headers(feature_set, config)

        return feature_set

    def _fix_issue_70(self, feature_set):
        """
        See issue #70: https://github.com/Dav1dde/glad/issues/70
        > it seems OSX already includes GLsizeiptr and a few others.
        > The same problem happens with glad.h as well.
        > The workaround appears to be to use long instead of ptrdiff_t.
        """
        for type_name in  ('GLsizeiptr', 'GLintptr', 'GLsizeiptrARB', 'GLintptrARB'):
            if type_name in feature_set.types:
                index = feature_set.types.index(type_name)
                type_ = copy.deepcopy(feature_set.types[index])
                type_._raw = \
                    '#if defined(__ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__) ' + \
                    '&& (__ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__ > 1060)\n' + \
                    type_._raw.replace('ptrdiff_t', 'long') + '\n#else\n' + type_._raw + '\n#endif'
                feature_set.types[index] = type_

    def _fix_cpp_style_comments(self, feature_set):
        """
        Turns CPP-Style comments `//` into C90 compatible comments `/**/`

        Currently the only "workaround" needed to make Vulkan compile with -ansi.
        See also: https://github.com/KhronosGroup/Vulkan-Docs/pull/700
        """
        for i, type_ in enumerate(feature_set.types):
            if '//' in type_._raw:
                new_type = copy.deepcopy(type_)
                new_type._raw = replace_cpp_style_comments(new_type._raw)
                feature_set.types[i] = new_type

    def _fixup_enums(self, feature_set):
        """
        There are some enums which are simply empty:
        https://github.com/KhronosGroup/Vulkan-Docs/issues/1754
        they need to be removed, we need to also remove any type which is an alias to that empty enum.

        Additionally we need to extend type information for enum alias types,
        if the alias points to an enum with bitwidth 64 copy over the
        bitwidth information so we can later produce the correct typedef.
        """
        bitwidth_64 = set()
        to_remove = set()

        for typ in (t for t in feature_set.types if isinstance(t, EnumType)):
            if typ.bitwidth == '64':
                bitwidth_64.add(typ.name)
            if typ.alias is None and not typ.enums_for(feature_set):
                to_remove.add(typ.name)

        new_types = []
        for typ in feature_set.types:
            if typ.alias:
                if typ.alias in bitwidth_64:
                    typ.bitwidth = '64'

            if typ.name not in to_remove and typ.alias not in to_remove:
                new_types.append(typ)

        feature_set.types = new_types

    def _replace_included_headers(self, feature_set, config):
        if not config['HEADER_ONLY']:
            return feature_set

        types = feature_set.types
        for header in self.ADDITIONAL_HEADERS:
            try:
                type_index = types.index(header.name)
            except ValueError:
                continue

            content = self._read_header(header.url)
            for pheader in self.ADDITIONAL_HEADERS:
                name = pheader.name.rsplit('/', 1)[-1]
                content = re.sub(
                    '^(#include\\s*["<]({}|{})(\\.h)?[>"])'.format(name, pheader.include), r'/* \1 */',
                    content,
                    flags=re.MULTILINE
                )

            types[type_index] = Type(header.name, raw=content)

    def _add_header_requirements(self, feature_set):
        added = set()

        for header in self.ADDITIONAL_HEADERS:
            if header.name not in feature_set.types:
                continue

            for require in (header.requires or []):
                if require not in added:
                    t = Type(require, raw="#include \"{}\"".format(require))
                    feature_set.types.insert(0, t)
                    added.add(require)

    def _add_additional_headers(self, feature_set, config):
        if config['HEADER_ONLY']:
            return

        for header in self.ADDITIONAL_HEADERS:
            if header.name not in feature_set.types:
                continue

            path = os.path.join(self.path, 'include/{}'.format(header.include))

            directory_path = os.path.split(path)[0]
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)

            if not os.path.exists(path):
                content = self._read_header(header.url)
                with open(path, 'wb') as dest:
                    dest.write(content.encode('utf-8'))

    def _read_header(self, url):
        if url not in self._headers:
            with closing(self.opener.urlopen(url)) as src:
                header = src.read().decode('utf-8')

            header = replace_cpp_style_comments(header)
            self._headers[url] = header

        return self._headers[url]




class ConfigException(Exception):
    pass


class InvalidConfig(ConfigException):
    pass


class OptionRequired(InvalidConfig):
    def __init__(self, name, option):
        ConfigException.__init__(self, 'Required option {!r} not set'.format(name))

        self.name = name
        self.option = option


class ConstraintException(InvalidConfig):
    def __init__(self, message, constraint):
        InvalidConfig.__init__(self, message)

        self.constraint = constraint


class RequirementNotSatisfied(ConstraintException):
    pass


class UnsupportedConfiguration(ConstraintException):
    pass


class InvalidOption(ConfigException):
    def __init__(self, name):
        ConfigException.__init__(self, 'Invalid option {!r}'.format(name))

        self.name = name


def identity(x):
    return x


def one_of(choices):
    def validator(x):
        if x not in choices:
            raise ValueError('invalid choice {!r}, expected one of {!r}', x, choices)

    validator.__config__doc__ = 'One of: {!r}'.format(choices)
    return validator


class ConfigOption(object):
    def __init__(self, description, converter=None, default=None, required=False, help=None):
        self.converter = converter
        if self.converter is None:
            self.converter = identity

        self.description = description
        self.default = default
        self.required = required
        self.help = help

        try:
            self.description = '{}. {}'.format(
                self.description.rstrip('. '),
                self.converter.__config_doc__
            )
        except AttributeError:
            pass

        if self.default and self.required:
            raise ValueError(
                'ConfigOption cannot have a default and be required at the same time.'
            )

    def to_parser_arguments(self):
        args = dict(
            type=self.converter,
            help=self.description,
        )

        if self.required:
            args['required'] = True
        else:
            if self.converter is bool:
                args['action'] = 'store_false' if self.default else 'store_true'
                args.pop('type')
            else:
                args['default'] = self.default

        return args


class Constraint(object):
    def validate(self, config):
        pass


class RequirementConstraint(Constraint):
    """
    Specifies a simple requirement constraint.
    If a list of options are given (True), the requirement needs to be True as well.
    This only checks for the variables boolean values.
    """
    def __init__(self, options, require, error_formatter=None):
        self.options = options
        self.require = require

        if len(self.options) == 0:
            raise ValueError('At least one option required')

        self.error_formatter = error_formatter
        if self.error_formatter is None:
            self.error_formatter = self._format_error

    def validate(self, config):
        if all(config[option] for option in self.options) and not config[self.require]:
            raise RequirementNotSatisfied(self.error_formatter(self, config), self)

    @staticmethod
    def _format_error(constraint, config):
        plural = len(constraint.options) > 1
        options = ', '.join(constraint.options[:-1])
        if options:
            options = '{} and {}'.format(options, constraint.options[-1])
        else:
            options = constraint.options[0]

        return 'option{s} {options} require{not_s} option {require}'.format(
            options=options,
            require=constraint.require,
            s='s' if plural else '',
            not_s='' if plural else 's'
        )


class UnsupportedConstraint(Constraint):
    """
    Specifies a unsupported constraint, that can either be a single option
    or a combination of options. Checks only for the boolean value of the options.
    """
    def __init__(self, given, not_allowed, error_formatter=None):
        self.given = given
        self.not_allowed = not_allowed

        if len(self.given) == 0:
            raise ValueError('At least one \'given\' option required')

        self.error_formatter = error_formatter
        if self.error_formatter is None:
            self.error_formatter = self._format_error

    def validate(self, config):
        if all(config[option] for option in self.given) and config[self.not_allowed]:
            raise UnsupportedConfiguration(self.error_formatter(self, config), self)

    @staticmethod
    def _format_error(constraint, config):
        plural = len(constraint.given) > 1
        given = ', '.join(constraint.given[:-1])
        if given:
            given = '{} and {}'.format(given, constraint.given[-1])
        else:
            given = constraint.given[-1]

        return 'option{s} {given} can not be used together with {not_allowed}'.format(
            given=given,
            not_allowed=constraint.not_allowed,
            s='s' if plural else ''
        )


class Config(object):
    """
    Base for all glad configurations. The class with initiliaze the options
    with it iself. Every uppercase name will be assumed to be a configuration option
    and should be of type ConfigOption:

        class MyAwesomeConfig(Config):
            DEBUG = ConfigOption(
                converter=bool,
                default=False
                description='Enables debug output'
            )
            ITERATIONS = ConfigOption(
                converter=int,
                required=True
                description='Number of iterations'
            )

        config = MyAwesomeConfig()
        config['DEBUG'] = True

        # update config from file
        # ...
        # now make sure every required option has been set
        config.validate()

        if config['DEBUG']:
            print 'debug information'

        Special Constraints can be specified in the __constraints__ variable.

            class MyConstraintConfig(MyAwesomeConfig):
                __constraints__ = [
                    RequirementConstraint(['DEBUG'], 'iterations')
                ]

        The constraints will be checked when calling `.valid` or `.validate()`
    """

    def __init__(self):
        self._options = dict()
        self._values = dict()

        # initialize options, every uppercase name without leading underscore = option
        for name in dir(self):
            if name.isupper() and not name.startswith('_'):
                option = self._options[name] = getattr(self, name)

                if not option.required:
                    self._values[name] = option.default

    def set(self, name, value, convert=True):
        try:
            option = self._options[name]
        except KeyError:
            raise InvalidOption(name)

        if convert:
            value = option.converter(value)
        self._values[name] = value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __getitem__(self, item):
        return self._values[item]

    def __setitem__(self, key, value):
        self.set(key, value, convert=True)

    def items(self):
        return list(self._options.items())

    @property
    def valid(self):
        """
        Checks if every required option has been set.

        :return: True if everything has been set otherwise False
        """
        try:
            self.validate()
        except InvalidConfig:
            return False
        return True

    def validate(self):
        """
        Checks if every required option has been set.
        Throws InvalidConfig if a required option is missing.
        This also checks all specified constraints.

        Should be overwritten by subclasses if necessary.
        """
        for name, option in self._options.items():
            if option.required:
                if not name in self._values:
                    raise OptionRequired(name, option)

        constraints = getattr(self, '__constraints__', [])
        for constraint in constraints:
            constraint.validate(self)

    def update_from_object(self, obj, convert=True, ignore_additional=False):
        for name in dir(obj):
            if not name.startswith('_'):
                try:
                    self.set(name, getattr(obj, name), convert=convert)
                except InvalidOption:
                    if not ignore_additional:
                        raise

    def init_parser(self, parser):
        for name, option in self._options.items():
            parser_name = '--' + name.lower().replace('_', '-')

            parser.add_argument(
                parser_name,
                dest=name,
                **option.to_parser_arguments()
            )

    def to_dict(self, transform=None):
        if transform is None:
            transform = identity

        result = dict()

        for name, value in self._values.items():
            result[transform(name)] = value

        return result



import os.path
import logging
import shutil

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

try:
    from importlib.resources import files

    def resource_open(package, name, *args, **kwargs):
        return files(package).joinpath(name).open(*args, **kwargs)
except ImportError:
    try:
        from pkg_resources import resource_stream

        def resource_open(package, name, *args, **kwargs):
            return resource_stream(package, name)
    except ImportError:
        def resource_open(package, name, *args, **kwargs):
            raise FileNotFoundError


BASE_PATH = os.path.abspath(os.path.dirname(__file__))

logger = logging.getLogger('glad.files')


class GladFileException(Exception):
    pass


def open_local(name, *args, **kwargs):
    # use pkg_resources when available, makes it work in zipped modules
    # or other environments
    try:
        return resource_open(__name__, name, *args, **kwargs)
    except FileNotFoundError:
        pass

    # fallback to filesystem
    logger.info('falling back to packaged path: %r', name)
    local_path = os.path.normpath(os.path.join(BASE_PATH, os.path.join(name)))
    if not local_path.startswith(BASE_PATH):
        raise GladFileException('unsafe file path, won\'t open {!r}'.format(local_path))
    return open(local_path, *args, **kwargs)


class StaticFileOpener(object):
    def urlopen(self, url, data=None, *args, **kwargs):
        logger.debug('intercepted attempt to retrieve resource: %r', url)
        if data is not None:
            raise GladFileException('can not resolve requests with payload')

        filename = urlparse(url).path.rsplit('/', 1)[-1]
        return open_local(filename, 'rb')

    def urlretrieve(self, url, filename, *args, **kwargs):
        with self.urlopen(url) as src:
            with open(filename, 'wb') as dst:
                shutil.copyfileobj(src, dst)



from glad.parse import Specification, Require


class EGL(Specification):
    DISPLAY_NAME = 'EGL'

    API = 'https://raw.githubusercontent.com/KhronosGroup/EGL-Registry/main/api/'
    NAME = 'egl'

    def protections(self, symbol, api=None, profile=None, feature_set=None):
        return list()


class GL(Specification):
    DISPLAY_NAME = 'OpenGL'

    API = 'https://raw.githubusercontent.com/KhronosGroup/OpenGL-Registry/main/xml/'
    NAME = 'gl'

    def _magic_require(self, api, profile):
        require = Specification._magic_require(self, api, profile)

        magic_blacklist = (
            'stddef', 'khrplatform', 'inttypes',  # gl.xml
        )
        requirements = [r for r in require.requirements if r not in magic_blacklist]
        return Require(api, profile, requirements)

    def protections(self, symbol, api=None, profile=None, feature_set=None):
        return list()


class GLX(Specification):
    DISPLAY_NAME = 'GLX'

    API = 'https://raw.githubusercontent.com/KhronosGroup/OpenGL-Registry/main/xml/'
    NAME = 'glx'

    def protections(self, symbol, api=None, profile=None, feature_set=None):
        return list()


class WGL(Specification):
    DISPLAY_NAME = 'WGL'

    API = 'https://raw.githubusercontent.com/KhronosGroup/OpenGL-Registry/main/xml/'
    NAME = 'wgl'

    def protections(self, symbol, api=None, profile=None, feature_set=None):
        return list()


class VK(Specification):
    DISPLAY_NAME = 'Vulkan'

    API = 'https://raw.githubusercontent.com/KhronosGroup/Vulkan-Docs/main/xml/'
    NAME = 'vk'

    def _magic_require(self, api, profile):
        # magic_categories = (
        #     'define', 'basetype', 'handle'
        # )
        #
        # requirements = [name for name, types in self.types.items()
        #                 if any(t.api in (None, api) and t.category in magic_categories for t in types)]
        #
        # return Require(api, profile, requirements)
        return None

    def _magic_are_enums_blacklisted(self, enums_element):
        # blacklist everything that has a type
        return enums_element.get('type') in ('enum', 'bitmask')
